import os.path

from dsdl.dataset import CheckDataset, ImageVisualizePipeline, Util, Report
import click
import json
from random import randint

try:
    from yaml import CSafeLoader as YAMLSafeLoader
except ImportError:
    from yaml import SafeLoader as YAMLSafeLoader
from .commons import OptionEatAll, prepare_input
from ..parser import check_dsdl_parser


@click.command(name="check")
@click.option("-y", "--yaml", "dsdl_yaml", type=str, required=True, help="the path of dsdl yaml file")
@click.option("-c", "--config", "config", type=str, required=True, help="the path of the config file")
@click.option("-l", "--location", "location", type=click.Choice(["local", "ali-oss"]), required=True,
              help="the path of the config file")
@click.option("-n", "--num", "num", type=int, default=5, help="how many samples sampled from the dataset")
@click.option("-r", "--random", is_flag=True, help="whether to sample randomly")
@click.option("-f", "--fields", cls=OptionEatAll, type=str, help="the task to visualize")
@click.option("-t", "--task", type=str, help="the task to visualize")
@click.option("-p", "--position", type=str, required=False, help='the directory of dsdl define file')
@click.option("-o", "--output", type=str, help="the dir to output the check report")
@prepare_input(visualize=True, multistage=False)
def check(dsdl_yaml, num, random, fields, config, position, output, **kwargs):
    assert os.path.isdir(output), "Please use '-o' to specify a existing working dir to generate the log files."
    log_dir = os.path.join(output, "log")
    os.makedirs(log_dir, exist_ok=True)
    report_obj = Report(os.path.join(log_dir, "output.md"))

    # parse
    if position:
        parse_report = check_dsdl_parser(dsdl_yaml["yaml_file"], dsdl_library_path=position, report_flag=True)
    else:
        parse_report = check_dsdl_parser(dsdl_yaml["yaml_file"], dsdl_library_path="", report_flag=True)
    dsdl_py = parse_report["dsdl_py"]
    parse_report = json.loads(parse_report["check_log"])
    report_obj.set_parser_info(parse_report)
    if parse_report['flag'] == 0:
        report_obj.generate()
        return
    exec(dsdl_py, {})

    dataset = CheckDataset(report_obj, dsdl_yaml["samples"], dsdl_yaml["sample_type"], config,
                           global_info_type=dsdl_yaml["global_info_type"], global_info=dsdl_yaml["global_info"])

    num = min(num, len(dataset))

    if not random:
        indices = list(range(num))
    else:
        indices = [randint(0, len(dataset) - 1) for _ in range(num)]

    samples = list()
    palette = dict()
    for ind in indices:
        samples.append(ImageVisualizePipeline(sample=dataset[ind], palette=palette, field_list=fields))

    print(Util.format_sample([s.format() for s in samples]))
    # 将读取的样本进行可视化
    for sample in samples:
        try:
            vis_sample = sample.visualize()
            for vis_name, vis_item in vis_sample.items():
                report_obj.add_image_info({"img": vis_item, "status": "success"})

        except Exception as e:
            report_obj.add_image_info({"img": None, "status": "fail", "msg": repr(e)})

    report_obj.generate()



import os.path
from environs import Env

local = dict(
    type="LocalFileReader",
    working_dir="local path of your media",
)

env = Env()
ali_oss = env.dict('ALI_OSS_DICT')


class Config:
    project_path = os.getcwd()
    root_path = os.path.join(project_path, "dsdl_test")
    temp_file_path = os.path.join(project_path, "temp.py")
    logger_path = os.path.join(project_path, "log_1030.txt")  # logger文件保存路径
    config_path = os.path.join(project_path, "config.py")   # dsdl view命令使用的配置文件路径
    v3_folder = os.path.join(root_path, "datasets")  # 存放v0.3格式的文件夹，会从oss下载到这个文件夹下
    dsdl_folder = os.path.join(
        root_path, "datasets_dsdl"
    )  # 存放dsdl格式的文件夹，从v0.3格式转换而来，与v0.3格式的文件夹不能相同
    unique_category = "category_name"  # 跑分类数据集的时候，可能需要修改此字段，详见分类转换脚本的readme
    dataset_list = {
        "classification": [
            "MNIST",
            # "UCF101",  #'VideoField'
            "Caltech-256",
            "Oxford_102_Flower",
            "Food-101",
            # "cats_vs_dogs",
            "STL-10",
            # "FGVC_Aircraft",
        ],
        "detection": [
            "FGVC_Aircraft",
            "KITTI_Object",
            # "ILSVRC2015DET",
            # "CityPersons",
            # "CrowdHuman",  # too large
            # "DOTA-v1.5",
            "xBD",
            "BDD100KImages",
        ],
    }




import click
from typing import Sequence, Union
import os

try:
    from yaml import CSafeLoader as YAMLSafeLoader
except ImportError:
    from yaml import SafeLoader as YAMLSafeLoader
from yaml import load as yaml_load
import json

TASK_FIELDS = {
    "detection": ["image", "label", "bbox", "polygon", "keypoint", "rotatedbbox"],
    "classification": ["image", "label"],
    "semantic-seg": ["image", "labelmap"],
    "panoptic-seg": ["image", "labelmap", "instancemap"],
    "ocr": ["image", "rotatedbbox", "text", "polygon"]
}


class OptionEatAll(click.Option):
    """
    implemented by https://stackoverflow.com/users/7311767/stephen-rauch in https://stackoverflow.com/questions/48391777/nargs-equivalent-for-options-in-click
    """

    def __init__(self, *args, **kwargs):
        self.save_other_options = kwargs.pop('save_other_options', True)
        nargs = kwargs.pop('nargs', -1)
        assert nargs == -1, 'nargs, if set, must be -1 not {}'.format(nargs)
        super(OptionEatAll, self).__init__(*args, **kwargs)
        self._previous_parser_process = None
        self._eat_all_parser = None

    def add_to_parser(self, parser, ctx):

        def parser_process(value, state):
            # method to hook to the parser.process
            done = False
            value = [value]
            if self.save_other_options:
                # grab everything up to the next option
                while state.rargs and not done:
                    for prefix in self._eat_all_parser.prefixes:
                        if state.rargs[0].startswith(prefix):
                            done = True
                    if not done:
                        value.append(state.rargs.pop(0))
            else:
                # grab everything remaining
                value += state.rargs
                state.rargs[:] = []
            value = tuple(value)

            # call the actual process
            self._previous_parser_process(value, state)

        retval = super(OptionEatAll, self).add_to_parser(parser, ctx)
        for name in self.opts:
            our_parser = parser._long_opt.get(name) or parser._short_opt.get(name)
            if our_parser:
                self._eat_all_parser = our_parser
                self._previous_parser_process = our_parser.process
                our_parser.process = parser_process
                break
        return retval


YAML_VALID_SUFFIX = ('.yaml', '.YAML')
JSON_VALID_SUFFIX = ('.json', '.JSON')
VALID_SUFFIX = YAML_VALID_SUFFIX + JSON_VALID_SUFFIX


def load_samples(dsdl_path: str, path: Union[str, Sequence[str]], extract_key="samples"):
    samples = []
    paths = []
    dsdl_dir = os.path.split(dsdl_path)[0]
    if isinstance(path, str):
        path = os.path.join(dsdl_dir, path)
        if os.path.isdir(path):
            paths = [os.path.join(path, _) for _ in os.listdir(path) if _.endswith(VALID_SUFFIX)]
        elif os.path.isfile(path):
            if path.endswith(VALID_SUFFIX):
                paths = [path]
    elif isinstance(path, (list, tuple)):
        paths = [os.path.join(dsdl_dir, _) for _ in path if os.path.isfile(_) and _.endswith(VALID_SUFFIX)]
    for p in paths:
        if p.endswith(YAML_VALID_SUFFIX):
            with open(p, "r") as f:
                data = yaml_load(f, YAMLSafeLoader)[extract_key]
            if isinstance(data, list):
                samples.extend(data)
            else:
                samples.append(data)
        else:
            with open(p, "r") as f:
                data = json.load(f)[extract_key]
            if isinstance(data, list):
                samples.extend(data)
            else:
                samples.append(data)
    return samples


def prepare_input(**kwargs):
    def _decorator(func):
        def process(dsdl_yaml, config, location, num, random, fields, task, position, **kwargs2):
            result = {
                "dsdl_yaml": dsdl_yaml,
                "config": config,
                "location": location,
                "num": num,
                "random": random,
                "fields": fields,
                "task": task,
                "position": position,
                **kwargs2
            }

            # parse samples and global-info
            with open(dsdl_yaml, "r") as f:
                dsdl_info = yaml_load(f, Loader=YAMLSafeLoader)['data']
                sample_type = dsdl_info['sample-type']
                global_info_type = dsdl_info.get("global-info-type", None)
                global_info = None
                if "sample-path" not in dsdl_info or dsdl_info["sample-path"] in ("local", "$local"):
                    assert "samples" in dsdl_info, f"Key 'samples' is required in {dsdl_yaml}."
                    samples = dsdl_info['samples']
                else:
                    sample_path = dsdl_info["sample-path"]
                    samples = load_samples(dsdl_yaml, sample_path)
                if global_info_type is not None:
                    if "global-info-path" not in dsdl_info:
                        assert "global-info" in dsdl_info, f"Key 'global-info' is required in {dsdl_yaml}."
                        global_info = dsdl_info["global_info"]
                    else:
                        global_info_path = dsdl_info["global-info-path"]
                        global_info = load_samples(dsdl_yaml, global_info_path, "global-info")[0]
            result["dsdl_yaml"] = {"samples": samples, "global_info": global_info, "sample_type": sample_type,
                                   "global_info_type": global_info_type, "yaml_file": dsdl_yaml}

            # parse location config
            config_dic = {}
            with open(config, encoding='utf-8') as config_file:
                exec(config_file.read(), config_dic)
            location_config = config_dic["local" if location == "local" else "ali_oss"]
            result["config"] = location_config

            # parse field list
            if task:
                assert task in TASK_FIELDS, f"invalid task, you can only choose in {list(TASK_FIELDS.keys())}"
                fields = TASK_FIELDS[task]
            else:
                if fields is None:
                    fields = []
                else:
                    local_dic = {}
                    exec(f"f = list({fields})", {}, local_dic)
                    fields = local_dic['f']
                fields = list(set(fields + ["image", "dict"]))
            fields = [_.lower() for _ in fields]

            result["fields"] = fields

            return result

        def wrapper(*args, **kwargs3):
            ret = process(*args, **kwargs3, **kwargs)
            ret = func(**ret)
            return ret

        return wrapper

    return _decorator



class DuplicateDefineWarning(SyntaxWarning):
    pass


class InvalidLabelWarning(SyntaxWarning):
    pass


class FieldNotFoundWarning(SyntaxWarning):
    pass


class ClassHasDefinedWarning(SyntaxWarning):
    pass


class ImportWarning(ImportWarning):
    pass


class DefineSyntaxWarning(SyntaxWarning):
    pass






from .local_studio_view import LocalStudioView
from .remote_studio_view import RemoteStudioView


def StudioView(dataset_name, task_type, n=None, shuffle=False, remote=False):
    if remote:
        return RemoteStudioView(dataset_name, task_type, n=n, shuffle=shuffle)
    else:
        return LocalStudioView(dataset_name, task_type, n=n, shuffle=shuffle)


__all__ = [
    "StudioView",
]



class BaseGeometry:

    def visualize(self, image, palette, **kwargs):
        return image

    def __repr__(self):
        return self.__class__.__name__ + " object"

    @property
    def field_key(self):
        return self.__class__.__name__



import os
import json
from boto3.session import Session
from collections import namedtuple
from ...objectio import AwsOSSFileReader
from .base_studio_view import BaseStudioView

s3_config = namedtuple("s3_config", ("ak", "sk", "endpoint", "region", "bucket"))
default_cfg = s3_config("ailabminio", "123123123", "http://10.140.0.94:9800", "ailab", "dsdldata")


class RemoteStudioView(BaseStudioView):

    def __init__(self, dataset_name, task_type, n=None, shuffle=False):
        self.remote_cfg = self._get_s3_config(dataset_name)
        os.system(f"odl-cli get {dataset_name} --label")
        super().__init__(dataset_name, task_type, n=n, shuffle=shuffle)

    def _init_file_reader(self):
        if self.remote_cfg is None:
            raise RuntimeError(f"remote dataset {self.dataset_name} not exist.")
        return AwsOSSFileReader(self.dataset_name,
                                self.remote_cfg.bucket,
                                self.remote_cfg.ak,
                                self.remote_cfg.sk,
                                self.remote_cfg.endpoint,
                                self.remote_cfg.region)

    def _get_s3_config(self, dataset_name):
        config_path = os.path.join(os.path.expanduser("~"), ".dsdl", "dsdl.json")
        with open(config_path, "r") as f:
            storage_info = json.load(f)["storage"]
        cfg_lst = [default_cfg]
        for storage in storage_info.values():
            if "ak" in storage and "sk" in storage and "endpoint" in storage:
                cfg = s3_config(storage["ak"], storage["sk"], storage["endpoint"],
                                storage.get("region", default_cfg.region), storage.get("bucket", default_cfg.bucket))
                cfg_lst.append(cfg)
        res = None
        for cfg in cfg_lst:
            session = Session(cfg.ak, cfg.sk)
            client = session.client("s3", endpoint_url=cfg.endpoint, region_name=cfg.region, use_ssl=False)
            if self.is_dataset_remote_exist(client, cfg.bucket, dataset_name):
                res = cfg
                break
        return res

    @staticmethod
    def is_dataset_remote_exist(client, bucket, dataset_name):

        child_dir_list = set()
        paginator = client.get_paginator('list_objects_v2')
        pages = paginator.paginate(Bucket=bucket, Prefix="", Delimiter='/')

        for page in pages:
            if 'CommonPrefixes' in page.keys():
                for p in page['CommonPrefixes']:
                    child_dir_list.add(p['Prefix'])

        dataset_list = [_.replace('/', '') for _ in child_dir_list]

        return dataset_name in dataset_list



from .local import LocalFileReader
from .ali_oss import AliOSSFileReader
from .base import BaseFileReader
from .ceph import CephFileReader, PetrelFileReader
from .aws_oss import AwsOSSFileReader

__all__ = [
    "LocalFileReader",
    "AliOSSFileReader",
    "BaseFileReader",
    "CephFileReader",
    "PetrelFileReader",
    "AwsOSSFileReader",
]



from .base_geometry import BaseGeometry


class Shape(BaseGeometry):
    def __init__(self, value, mode, media):
        media = media.lower()
        mode = mode.lower()
        self._media_type = media
        self._value = value
        self._mode = mode

    @property
    def media_type(self):
        return self._media_type

    @property
    def value(self):
        return self._value

    @property
    def mode(self):
        return self._mode

    def __repr__(self):
        return f"{self.media_type} shape: {self.value}"


class ImageShape(Shape):
    def __init__(self, value, mode="hw"):
        assert mode.lower() in ("hw", "wh")
        super(ImageShape, self).__init__(value, mode.lower(), "image")
        if mode == "hw":
            self._height, self._width = value[0], value[1]
        else:
            self._width, self._height = value[0], value[1]

    @property
    def height(self):
        return self._height

    @property
    def width(self):
        return self._width

    def __repr__(self):
        return f"image height: {self.height}; image width: {self.width}"

    @property
    def field_key(self):
        return "ImageShape"



from .base_geometry import BaseGeometry
from PIL import ImageDraw, ImageFont
import numpy as np


class Text(BaseGeometry):
    def __init__(self, text):
        self._text = text

    @property
    def value(self):
        return self._text

    @property
    def text(self):
        return self._text

    def visualize(self, image, palette, **kwargs):
        draw_obj = ImageDraw.Draw(image)
        text_color = (0, 255, 0)  # green
        font = ImageFont.load_default()
        label_size = draw_obj.textsize(self.value, font)
        if "bbox" in kwargs:
            coords = np.array([[item.xyxy[0], item.xyxy[3] - 1.2 * label_size[1]] for item in kwargs["bbox"].values()])
        elif "polygon" in kwargs:
            coords = np.array(
                [[item.point_for_draw("lb")[0], item.point_for_draw("lb")[1] - 1.2 * label_size[1]] for item in
                 kwargs["polygon"].values()])
        elif "rotatedbbox" in kwargs:
            coords = np.array(
                [[item.point_for_draw("lb")[0], item.point_for_draw("lb")[1] - 1.2 * label_size[1]] for item in
                 kwargs["rotatedbbox"].values()])
        else:
            coords = np.array([[0, image.size[0] - 1.2 * label_size[1]]])
        for coord in coords:
            draw_obj.rectangle([tuple(coord), tuple(coord + label_size)], fill=(*text_color, 255))
            draw_obj.text(tuple(coord), self.value, fill=(255, 255, 255, 255), font=font)
        del draw_obj
        return image

    def __repr__(self):
        return self._text

    @property
    def field_key(self):
        return "Text"



from typing import List, overload
import numpy as np
from PIL import ImageDraw, Image
from .base_geometry import BaseGeometry


class PolygonItem(BaseGeometry):

    def __init__(
            self,
            points: List[List[float]]
    ):
        self._data = points

    @property
    def points(self) -> List[List[float]]:
        return self._data

    @property
    def points_x(self) -> List[float]:
        return [_[0] for _ in self._data]

    @property
    def points_y(self) -> List[float]:
        return [_[1] for _ in self._data]

    def point_for_draw(self, mode: str = "lt"):
        assert mode in ("lb", "lt", "rb", "rt")
        if mode == "lt":
            p = min(self._data, key=lambda x: x[0] + x[1])
        elif mode == "lb":
            p = min(self._data, key=lambda x: x[0] - x[1])
        elif mode == "rt":
            p = min(self._data, key=lambda x: -x[0] + x[1])
        else:  # rb
            p = min(self._data, key=lambda x: -x[0] - x[1])
        p = [int(_) for _ in p]
        return p

    @property
    def openmmlabformat(self) -> List[float]:
        return self._flatten()

    def to_tuple(self):
        return tuple([(_[0], _[1]) for _ in self._data])

    def _flatten(self) -> List[float]:
        return [_ for point in self._data for _ in point]

    def __repr__(self):
        return str(self._data)


class Polygon(BaseGeometry):

    def __init__(self, polygons: List[PolygonItem]):
        self._data = polygons

    @property
    def polygons(self):
        return self._data

    @property
    def openmmlabformat(self) -> List[List[float]]:
        return [_.openmmlabformat for _ in self._data]

    def point_for_draw(self, mode: str = "lt") -> [int, int]:
        assert mode in ("lb", "lt", "rb", "rt")
        if mode == "lt":
            p = min(self._data, key=lambda x: x.point_for_draw(mode)[0] + x.point_for_draw(mode)[1])
        elif mode == "lb":
            p = min(self._data, key=lambda x: x.point_for_draw(mode)[0] - x.point_for_draw(mode)[1])
        elif mode == "rt":
            p = min(self._data, key=lambda x: -x.point_for_draw(mode)[0] + x.point_for_draw(mode)[1])
        else:  # rb
            p = min(self._data, key=lambda x: -x.point_for_draw(mode)[0] - x.point_for_draw(mode)[1])
        p = [int(_) for _ in p.point_for_draw(mode)]
        return p

    def visualize(self, image, palette, **kwargs):
        color = (0, 255, 0)
        if "label" in kwargs:
            for label in kwargs["label"].values():
                if label.category_name not in palette:
                    palette[label.category_name] = tuple(np.random.randint(0, 255, size=[3]))

                color = palette[label.category_name]

        poly = Image.new('RGBA', image.size[:2])
        pdraw = ImageDraw.Draw(poly)
        for polygon_item in self.polygons:
            pdraw.polygon(polygon_item.to_tuple(), fill=(*color, 127), outline=(*color, 255))
        image.paste(poly, mask=poly)
        del pdraw
        return image

    def __repr__(self):
        return str(self._data)

    @property
    def field_key(self):
        return "Polygon"


class RLEPolygon(BaseGeometry):
    def __init__(self, rle_data, image_shape):
        """
        rle_data: rle list
        image_shape: [H, W]
        """
        self._rle_data = rle_data
        self._image_shape = image_shape

    @property
    def rle_data(self):
        return self._rle_data

    @property
    def image_shape(self):
        return self._image_shape

    @property
    def mask(self):
        '''
        # ref: https://www.kaggle.com/paulorzp/run-length-encode-and-decode
        mask_rle: run-length as string formated (start length)
        shape: (height,width) of array to return
        Returns numpy array, 1 - mask, 0 - background
        '''
        s = self._rle_data
        shape = self._image_shape
        starts, lengths = [np.asarray(x, dtype=int) for x in (s[0:][::2], s[1:][::2])]
        starts -= 1
        ends = starts + lengths
        img = np.zeros(shape[0] * shape[1], dtype=np.uint8)
        for lo, hi in zip(starts, ends):
            img[lo:hi] = 1
        return img.reshape(shape)

    @property
    def openmmlabformat(self):
        return {"counts": self.rle_data, "size": self.image_shape}

    def point_for_draw(self, mode: str = "lt") -> [int, int]:
        pass

    def visualize(self, image, palette, **kwargs):
        color = (0, 255, 0)
        mask = self.mask
        color_seg = np.array((mask.shape[0], mask.shape[1], 3), dtype=np.uint8)

        if "label" in kwargs:
            for label in kwargs["label"].values():
                if label.category_name not in palette:
                    palette[label.category_name] = tuple(np.random.randint(0, 255, size=[3]))

                color = palette[label.category_name]
        color_seg[mask == 1, :] = np.array(color)
        overlay = Image.fromarray(color_seg).convert("RGBA")
        overlayed = Image.blend(image, overlay, 0.5)
        return overlayed



from fastapi import APIRouter
from dsdl.server.scheme import BaseResponse, CheckParam
from dsdl.parser import dsdl_parse, CHECK_LOG
from typing import List
import json

router = APIRouter()


@router.post("/parse_yaml")
def parse_and_check_yaml(
        param: CheckParam
) -> List[dict]:
    """
        该接口是在输入各种条件以后返回所有的该数据库符合要求的数据的页数
    Args:\n
        param:\n
            dsdl_yaml: 数据的yaml文件，一般是train.yaml、test.yaml(必须)
            dsdl_library_path: 其他yaml存放路径，默认的其他yaml存放路径是`dsdl/dsdl_library`
            output_file: 生成.py文件路径，默认None，以字符串格式返回
            report_flag: 是否生成报告json，是的话会在"check_log"中返回
    """
    dsdl_py = dsdl_parse(dsdl_yaml=param.dsdl_yaml, dsdl_library_path=param.dsdl_library_path,
                         output_file=param.output_file, report_flag=param.check_flag)
    res = {"dsdl_py": dsdl_py}
    if param.check_flag:
        res["check_log"] = json.dumps(CHECK_LOG.to_struct())
    return BaseResponse.decorate(res)







from .box import BBox
from .label import Label, LabelList
from .media import ImageMedia
from .polygon import Polygon, PolygonItem
from .attrbutes import Attributes
from .segmap import SegmentationMap
from .insmap import InstanceMap
from .keypoint import Coord2D, KeyPoints
from .registry import STRUCT, CLASSDOMAIN, LABEL
from .class_domain import ClassDomain
from .text import Text
from .rotate_box import RBBox
from .shape import Shape, ImageShape
from .instanceid import InstanceID

__all__ = [
    "BBox",
    "Label",
    "Text",
    "ImageMedia",
    "LabelList",
    "Polygon",
    "PolygonItem",
    "Attributes",
    "SegmentationMap",
    "InstanceMap",
    "Coord2D",
    "STRUCT",
    "CLASSDOMAIN",
    "LABEL",
    "ClassDomain",
    "KeyPoints",
    "RBBox",
    "Shape",
    "ImageShape",
    "InstanceID",
]



"""

cd function

Examples:
    $python cli.py studio coco
    Collecting usage statistics. To deactivate, set browser.gatherUsageStats to False.
    You can now view your Streamlit app in your browser.
    Network URL: http://127.0.0.1:8501
    External URL: http://?.?.?.?:8501
"""

import sys

from utils.views.view import View
from commands.cmdbase import CmdBase
from commons.exceptions import CLIException, ExistCode
from commons.argument_parser import EnvDefaultVar
from utils.admin import DBClient
from commands.const import (
    DEFAULT_CLI_CONFIG_FILE,
    DEFAULT_CLI_LOG_FILE_PATH,
    DEFAULT_CONFIG_DIR,
    DEFAULT_LOCAL_STORAGE_PATH,
    PROG_NAME,
    SQLITE_DB_PATH,
)

import commons.stdio as stdio


class Studio(CmdBase):
    """
    cd command
    """

    def init_parser(self, subparsers):
        """
        Initialize the parser for the command
        document : https://docs.python.org/zh-cn/3/library/argparse.html#

        Args:
            subparsers:

        Returns:

        """
        studio_parser = subparsers.add_parser(
            "studio",
            help="view the specified dataset on webpage.",
            example="studio.example",
        )

        studio_parser.add_argument(
            "dataset_name",
            action=EnvDefaultVar,
            envvar="DSDL_CLI_DATASET_NAME",
            type=str,
            help="dataset name",
            metavar="[DATASET NAME]",
        )
        studio_parser.add_argument(
            "-t",
            "--task",
            required=False,
            choices=["detection", "classification", "semantic-seg", "panoptic-seg"],
            help="task type [detection, classification, semantic-seg, panoptic-seg] of the dataset. ",
            metavar="[TASK TYPE]",
        )
        studio_parser.add_argument(
            "-n",
            "--number",
            required=False,
            type=int,
            help="number of images want to be viewed of the dataset.",
        )

        studio_parser_group = studio_parser.add_mutually_exclusive_group()

        studio_parser_group.add_argument(
            "-l",
            "--local",
            action="store_true",
            default=True,
            help="view local [ default ] dataset.",
        )
        studio_parser_group.add_argument(
            "-r",
            "--remote",
            action="store_true",
            default=False,
            help="view remote dataset.",
        )

        return studio_parser

    def cmd_entry(self, cmdargs, config, *args, **kwargs):
        """
        Entry point for the command

        Args:
            self:
            cmdargs:
            config:

        Returns:

        """

        dataset_name = cmdargs.dataset_name
        task_type_name = cmdargs.task
        number = cmdargs.number
        local = cmdargs.local
        remote = cmdargs.remote

        dbcli = DBClient()
        local_exists = dbcli.is_dataset_local_exist(dataset_name)
        local_media_exists = dbcli.is_dataset_media_downloaded(dataset_name)

        # TODO support remote dataset exists or not in the future:
        # remote_exists = dbcli.is_dataset_remote_exist(dataset_name)

        if task_type_name is None:
            # view = View(dataset_name, task=False)
            dataset_info = dbcli.get_dataset_info_by_name(dataset_name)
            if dataset_info is not None:
                task_type_name = dataset_info["task_type"]
                task_type_name_map = {
                    "ObjectDetection": "detection",
                    "Image Classification": "classification",
                    # "semantic-seg": "Segmentation", # TODO?
                    # "panoptic-seg": "Segmentation",
                }
                task_type_name = task_type_name_map[task_type_name]

            else:
                pass  # TODO raise?

        if task_type_name not in [
            "detection",
            "classification",
            "semantic-seg",
            "panoptic-seg",
        ]:
            raise CLIException(
                ExistCode.VIEW_TASK_TYPE_ERROR,
                f"Task type {task_type_name} is not supported.",
            )
        else:
            view = View(
                dataset_name,
                task=True,
                task_type=task_type_name,
                number=number,
            )

        # print(f"local:{local}")
        # print(f"remote:{str(remote)}")
        # print(f"local_exists: {local_exists}")

        if remote is True:
            if local_exists is True and local_media_exists is True:
                use_local = input(
                    f"Dataset {dataset_name} already exists on local.\nDo you want view using local dataset:(y/n)? "
                )
                if use_local in ["y", "Y", "yes", "Yes"]:
                    view.view_local_dataset()
                elif use_local in ["n", "N", "no", "No"]:
                    continue_remote = input(
                        f"Do you want continue viewing remote dataset:(y/n)? "
                    )
                    if continue_remote in ["y", "Y", "yes", "Yes"]:
                        view.view_remote_dataset()
                    elif continue_remote in ["n", "N", "no", "No"]:
                        stdio.print_stdout(msg="Visualization terminated.")
                        sys.exit(0)
            else:
                stdio.print_stderr(
                    "Dataset not exists on local,try to view remote dataset."
                )
                view.view_remote_dataset()
        elif local is True:
            if local_exists is True:
                view.view_local_dataset()
            else:
                stdio.print_stderr(f"Dataset {dataset_name} does not exist locally.")



import argparse
import os

import streamlit as st
import matplotlib.pyplot as plt
from PIL import Image

from dsdl.tools import StudioView


def main():
    parser = argparse.ArgumentParser(description="streamlit app inside odl-cli.")

    parser.add_argument(
        "--dataset-name",
    )
    parser.add_argument(
        "--task-type",
    )
    parser.add_argument(
        "--number",
        type=int,
    )
    parser.add_argument(
        "--remote",
        action="store_true",
    )

    try:
        args = parser.parse_args()
    except SystemExit as e:
        os._exit(e.code)

    dataset_name = args.dataset_name
    task_type = args.task_type
    number = args.number
    remote = args.remote

    if remote:
        # ex. iterator = StudioView("CIFAR-10", "classification", n=10, shuffle=True, remote=True)
        iterator = StudioView(
            dataset_name, task_type, n=number, shuffle=True, remote=True
        )
    else:
        # ex. iterator = StudioView("CIFAR-10", "classification", n=10, shuffle=True)
        iterator = StudioView(dataset_name, task_type, n=number, shuffle=True)

    image_list = []
    i = 0
    for image in iterator:
        if i == number:
            break
        else:
            image_list.append(image)
        i += 1

    st.title("Files")

    # display_images(image_list, max_images=number)
    image_grid_main(image_list, number)


def image_grid_main(images: [Image], max_images):
    st.title("Image Grid Display")
    if not images:
        print("No images to display.")
        return
    if len(images) > max_images:
        print(f"Showing {max_images} images of {len(images)}:")
        images = images[0:max_images]
    n = st.number_input("Select Grid Width", 1, 5, 3)

    groups = []
    for i in range(0, len(images), n):
        groups.append(images[i : i + n])

    for group in groups:
        cols = st.columns(n)
        for i, image_file in enumerate(group):
            cols[i].image(image_file)


if __name__ == "__main__":
    main()



import pytest
from dataclasses import dataclass, field
from click.testing import CliRunner
from glob import glob
from typing import List, TextIO, Union
import os
import oss2
from enum import Enum
from tunas2dsdl.cli import convert
from tunas2dsdl_c.dsdl_yaml_script import ConvertV3toDsdlYaml
from dsdl.parser import dsdl_parse
from dsdl.tools import view
from pylint import epylint as lint
from dsdl.test.config import Config, ali_oss
import pathlib
from collections import defaultdict
import shutil
import sys

config = Config()


class TaskEnum(Enum):
    CLS = "classification"
    DETE = "detection"
    # SEG = "segmentation"
    # KP = "keypoints"


@dataclass()
class DsdlTask:
    task_name: TaskEnum.CLS
    data_tunas_path: str
    data_dsdl_path: str
    data_name_list: List[str] = field(default_factory=list)

    def __post_init__(self):
        self.test_dir_list = []
        for dataset_name in self.data_name_list:
            dst = os.path.join(self.data_tunas_path, dataset_name)
            dst_dsdl = os.path.join(self.data_dsdl_path, dataset_name)
            self.test_dir_list.append((dst, dst_dsdl))


TaskData = defaultdict(DsdlTask)
for task in TaskEnum:
    TaskData[task.name] = DsdlTask(
        task_name=task.value,
        data_tunas_path=os.path.join(config.v3_folder, task.value),
        data_dsdl_path=os.path.join(config.dsdl_folder, task.value),
        data_name_list=config.dataset_list[task.value],
    )


def _validate_view(yaml_path: str, task_type: TaskEnum, logger: TextIO):
    logger.write(f"Test view {yaml_path}: \n")
    runner = CliRunner()
    if task_type == TaskEnum.DETE:
        result = runner.invoke(
            view,
            [
                "-y",
                yaml_path,
                "-c",
                config.config_path,
                "-l",
                "ali-oss",
                "-n",
                5,
                "-f",
                "Label BBox Polygon",
                "-p",
                os.path.dirname(yaml_path),
            ],
        )
    elif task_type == TaskEnum.CLS:
        result = runner.invoke(
            view,
            [
                "-y",
                yaml_path,
                "-c",
                config.config_path,
                "-l",
                "ali-oss",
                "-n",
                2,
                "-f",
                "Label",
                "-p",
                os.path.dirname(yaml_path),
            ],
        )
    else:
        raise TypeError(f"Unsupported task type: {task_type}.")
    if result.exit_code == 0:
        logger.write("success.\n")
        # logger.write(f"{result.output}\n")
    else:
        err_msg = f"Error in view code task: {yaml_path})."
        logger.write(f"{err_msg}\n\n")
        raise Exception(err_msg)


def test_python_file(temp_file: Union[str, pathlib.Path], logger: TextIO):
    (pylint_stdout, pylint_stderr) = lint.py_run(
        f"{temp_file} --errors-only", return_std=True
    )  # --msg-template={msg_id}
    if len(pylint_stdout.getvalue()) != 0:
        logger.write(f"{pylint_stdout.getvalue()}\n")
        raise Exception(pylint_stdout.getvalue())


def _validate_parser(yaml_path: Union[str, pathlib.Path], logger: TextIO):
    logger.write(f"Test parse {yaml_path}: \n")
    try:
        temp_file_path = config.temp_file_path
        library_path = os.path.dirname(yaml_path)
        dsdl_parse(yaml_path, library_path, temp_file_path)
        test_python_file(temp_file_path, logger)
        os.remove(temp_file_path)
        logger.write(f"success\n")
    except Exception:
        err_msg = f"Error in parser code task: {yaml_path}."
        logger.write(f"{err_msg}\n\n")
        raise Exception(err_msg)


def _percentage(consumed_bytes, total_bytes):
    """下载的进度条"""
    if total_bytes:
        rate = int(100 * (float(consumed_bytes) / float(total_bytes)))
        print("\r{0}% ".format(rate), end="")
        sys.stdout.flush()


def _download(bucket, src, dst):
    """下载函数"""
    print(f"Downloading {os.path.basename(dst)}...")

    object_stream = bucket.get_object(src, progress_callback=_percentage)
    with open(dst, "wb") as local_fileobj:
        shutil.copyfileobj(object_stream, local_fileobj)

    if object_stream.client_crc != object_stream.server_crc:
        print(f"The CRC checksum between client and server is inconsistent: {dst}")
        raise Exception(
            f"The CRC checksum between client and server is inconsistent: {dst}"
        )
    else:
        print("Done.")


class TestPipe:
    def setup_class(self):
        print("\nhere is setup_class")
        if os.path.isfile(config.logger_path):
            os.remove(config.logger_path)
        self.logger = open(config.logger_path, "a")
        auth = oss2.Auth(
            access_key_id=ali_oss["access_key_id"],
            access_key_secret=ali_oss["access_key_secret"],
        )
        self.bucket = oss2.Bucket(
            auth, endpoint=ali_oss["endpoint"], bucket_name=ali_oss["bucket_name"]
        )
        for task_name in TaskEnum:
            for dataset_name in TaskData[task_name.name].data_name_list:
                self.logger.write(f"========= {dataset_name} =========\n")
                src = os.path.join(dataset_name, "standard/0.3/")
                dst_tunas = os.path.join(
                    TaskData[task_name.name].data_tunas_path, dataset_name
                )
                # 如果存放v0.3格式的文件夹或存放dsdl格式的文件夹存在，也会先删除
                if not os.path.exists(dst_tunas):
                    os.makedirs(os.path.join(dst_tunas, "annotations/json"))
                    try:
                        _download(
                            self.bucket,
                            os.path.join(src, "dataset_info.json"),
                            os.path.join(dst_tunas, "dataset_info.json"),
                        )
                        for obj in oss2.ObjectIterator(
                            self.bucket,
                            prefix=os.path.join(src, "annotations/json/"),
                            delimiter="/",
                        ):
                            if obj.key.endswith("/"):
                                continue
                            _download(
                                self.bucket,
                                obj.key,
                                os.path.join(
                                    dst_tunas, "/".join(obj.key.split("/")[-3:])
                                ),
                            )
                        self.logger.write(f"success.\n")
                    except Exception as err:
                        print(err)
                        self.logger.write(f"{err}\n")
                        continue
                else:
                    self.logger.write(f"{dataset_name} v3 data already downloaded\n")

    @pytest.mark.parametrize(
        "tunas_dir, dsdl_dir", TaskData[TaskEnum.CLS.name].test_dir_list
    )
    def test_cls_tunas2dsdl(self, tunas_dir, dsdl_dir):
        self.logger.write(f"Test {TaskEnum.CLS.value} tunas2dsdl\n")
        try:
            self.logger.write(f"convert {os.path.basename(tunas_dir)}: \n")
            v3toyaml = ConvertV3toDsdlYaml(dataset_path=tunas_dir, output_path=dsdl_dir)
            v3toyaml.convert_pipeline()
        except Exception:
            err_msg = f"Error in tunas to dsdl code ({TaskEnum.CLS.value} task)."
            self.logger.write(f"{err_msg}\n\n")
            raise Exception(err_msg)
        else:
            self.logger.write("success.\n\n")

    @pytest.mark.parametrize(
        "tunas_dir, dsdl_dir", TaskData[TaskEnum.DETE.name].test_dir_list
    )
    def test_det_tunas2dsdl(self, tunas_dir, dsdl_dir):
        self.logger.write(f"Test {TaskEnum.DETE.value} tunas2dsdl\n")
        if not os.path.exists(dsdl_dir):
            os.makedirs(dsdl_dir)
        runner = CliRunner()
        result = runner.invoke(
            convert,
            ["-d", str(tunas_dir), "-w", str(dsdl_dir), "-t", "detection", "-s"],
        )
        if result.exit_code == 0:
            self.logger.write("success.\n")
            self.logger.write(result.output)
        else:
            self.logger.write(f"failed: ")
            err_msg = f"Error in tunas to dsdl code ({TaskEnum.DETE.value}: {os.path.basename(tunas_dir)})."
            self.logger.write(f"{err_msg}\n")
            raise Exception(err_msg)

    @pytest.mark.skip(reason="tunas2dsdl seg api is unfinished model")
    def test_seg_tunas2dsdl(self):
        for tunas_dir, dsdl_dir in self.SEG_TASK.test_dir_list:
            runner = CliRunner()
            result = runner.invoke(
                convert,
                ["-d", str(tunas_dir), "-w", str(dsdl_dir), "-t", "detection", "-s"],
            )
            if result.exit_code == 0:
                self.logger.write("success.")
                self.logger.write(result.output)
            else:
                self.logger.write("failed.")

    @pytest.mark.parametrize(
        "_, dsdl_folder", TaskData[TaskEnum.CLS.name].test_dir_list
    )
    def test_parser_cls_dsdl(self, _, dsdl_folder):
        for file in glob(os.path.join(dsdl_folder, "*_data.yaml")):
            _validate_parser(file, self.logger)

    @pytest.mark.parametrize(
        "_, dsdl_folder", TaskData[TaskEnum.DETE.name].test_dir_list
    )
    def test_parser_det_dsdl(self, _, dsdl_folder):
        for seg_folder in os.listdir(dsdl_folder):
            if os.path.isdir(os.path.join(dsdl_folder, seg_folder)):
                file = os.path.join(dsdl_folder, seg_folder, seg_folder + ".yaml")
                if os.path.exists(file):
                    _validate_parser(file, self.logger)

    @pytest.mark.parametrize(
        "_, dsdl_folder", TaskData[TaskEnum.CLS.name].test_dir_list
    )
    def test_view_cls_dsdl(self, _, dsdl_folder):
        for file in glob(os.path.join(dsdl_folder, "*_data.yaml")):
            _validate_view(file, TaskEnum.CLS, self.logger)

    @pytest.mark.parametrize(
        "_, dsdl_folder", TaskData[TaskEnum.DETE.name].test_dir_list
    )
    def test_view_det_dsdl(self, _, dsdl_folder):
        for seg_folder in os.listdir(dsdl_folder):
            if os.path.isdir(os.path.join(dsdl_folder, seg_folder)):
                file = os.path.join(dsdl_folder, seg_folder, seg_folder + ".yaml")
                if os.path.exists(file):
                    _validate_view(file, TaskEnum.DETE, self.logger)

    def teardown_class(self):

        self.logger.close()


import uvicorn
from dsdl.server.routers import api_parser


def create_fast_api_app(cors: bool = True, health_check: bool = True, **kwargs):
    from fastapi import FastAPI
    from starlette.middleware.cors import CORSMiddleware

    app = FastAPI(**kwargs)
    if cors:
        # 跨域
        app.add_middleware(
            CORSMiddleware,
            allow_origins=["*"],
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )

    if health_check:
        @app.get("/health-check")
        async def health_check_view():
            return {"code": "0000", "msg": "health"}

    return app


app = create_fast_api_app(
    title="dsdl checkout server",
    version="0.0.0",
    description="bankend of dsdl",
    openapi_prefix="",
)

# @app.get("/")
# async def root():
#     return {"message": "Hello World"}


app.include_router(api_parser.router, prefix="/parser", tags=["parser api"])


if __name__ == "__main__":
    uvicorn.run(app, host="127.0.0.1", port=8080)



from ..dataset import Dataset, ImageVisualizePipeline, Util
import click
import numpy as np
from random import randint
import cv2
import os

try:
    from yaml import CSafeLoader as YAMLSafeLoader
except ImportError:
    from yaml import SafeLoader as YAMLSafeLoader
from .commons import OptionEatAll, prepare_input
from ..parser import dsdl_parse


@click.command(name="view")
@click.option("-y", "--yaml", "dsdl_yaml", type=str, required=True, help="the path of dsdl yaml file")
@click.option("-c", "--config", "config", type=str, required=True, help="the path of the config file")
@click.option("-l", "--location", "location", type=click.Choice(["local", "ali-oss"]), required=True,
              help="the path of the config file")
@click.option("-n", "--num", "num", type=int, default=5, help="how many samples sampled from the dataset")
@click.option("-r", "--random", is_flag=True, help="whether to sample randomly")
@click.option("-v", "--visualize", is_flag=True, help="whether to visualize the sample selected")
@click.option("-f", "--fields", cls=OptionEatAll, type=str, help="the task to visualize")
@click.option("-t", "--task", type=str, help="the task to visualize")
@click.option("-p", "--position", type=str, required=False, help='the directory of dsdl define file')
@click.option("-m", "--multistage", is_flag=True, help="whether to use the generated python file")
@prepare_input(output=None)
def view(dsdl_yaml, num, random, visualize, fields, config, position, multistage, **kwargs):
    # parse
    if multistage:
        dsdl_py = os.path.splitext(dsdl_yaml["yaml_file"])[0] + ".py"
        with open(dsdl_py, encoding='utf-8') as dsdl_file:
            exec(dsdl_file.read(), {})
    else:
        if position:
            dsdl_py = dsdl_parse(dsdl_yaml["yaml_file"], dsdl_library_path=position)
        else:
            dsdl_py = dsdl_parse(dsdl_yaml["yaml_file"], dsdl_library_path="")
        exec(dsdl_py, {})

    dataset = Dataset(dsdl_yaml["samples"], dsdl_yaml["sample_type"], config,
                      global_info=dsdl_yaml["global_info"], global_info_type=dsdl_yaml["global_info_type"])

    num = min(num, len(dataset))
    if not random:
        indices = list(range(num))
    else:
        indices = [randint(0, len(dataset) - 1) for _ in range(num)]

    samples = list()
    palette = dict()
    for ind in indices:
        samples.append(ImageVisualizePipeline(sample=dataset[ind], palette=palette, field_list=fields))

    print(Util.format_sample([s.format() for s in samples]))
    # 将读取的样本进行可视化
    if visualize:
        for sample in samples:
            vis_sample = sample.visualize()
            for vis_name, vis_item in vis_sample.items():
                cv2.imshow(vis_name, cv2.cvtColor(np.array(vis_item), cv2.COLOR_BGR2RGB))
                cv2.waitKey(0)



"""
一个命令的实现例子

Examples:
    $python cli.py example --show tell me the truth
    >> Namespace(show=["'tell", 'me', 'the', "truth'"], command_handler=<bound method Example.cmd_entry of <commands.example.Example object at 0x0000017E6FD1DB40>>)
    >> ["'tell", 'me', 'the', "truth'"]
"""
import os

from commands.cmdbase import CmdBase
from commands.const import DSDL_CLI_DATASET_NAME
from commons.argument_parser import EnvDefaultVar
from tabulate import tabulate
from rich.tree import Tree
from rich import print as rprint
from utils import admin, query, plot
from utils.oss_ops import ops
import yaml
from commons.exceptions import CLIException, ExistCode
from commons.stdio import print_stdout
from loguru import logger

aws_access_key_id = query.aws_access_key_id
aws_secret_access_key = query.aws_secret_access_key
endpoint_url = query.endpoint_url
region_name = query.region_name
default_bucket = query.default_bucket


class Info(CmdBase):
    """
    Example command
    """

    def init_parser(self, subparsers):
        """
        Initialize the parser for the command
        document : https://docs.python.org/zh-cn/3/library/argparse.html#

        Args:
            subparsers:

        Returns:

        """
        inspect_parser = subparsers.add_parser(
            "info",
            help="inspect dataset info",
            example="info.example",
            description="inspect dataset info",
        )  # example 样例文件位于resources/下，普通的文本文件，每个命令写一个

        inspect_parser.add_argument(
            "dataset_name",
            action=EnvDefaultVar,
            envvar=DSDL_CLI_DATASET_NAME,
            type=str,
            help="dataset name. The arg is optional only when the default dataset name was set by cd command.",
        )

        # inspect_parser.add_argument(
        #     "--split",
        #     type=str,
        #     help="the split name of the dataset, such as train/test/validation split.",
        #     metavar="",
        # )

        # group = inspect_parser.add_mutually_exclusive_group()

        # group.add_argument(
        #     "-d",
        #     "--description",
        #     action="store_true",
        #     help="split name of the dataset, such as train/test/unlabeled or user self-defined split.",
        # )
        # group.add_argument(
        #     "-s",
        #     "--statistics",
        #     action="store_true",
        #     help="some statistics of the dataset.",
        # )
        # group.add_argument(
        #     "-m",
        #     "--metadata",
        #     action="store_true",
        #     help="show metadata of the dataset.",
        # )
        # group.add_argument(
        #     "--schema",
        #     action="store_true",
        #     help="show schema of the dataset.",
        # )
        # group.add_argument(
        #     "--preview",
        #     action="store_true",
        #     help="preview of the dataset.",
        # )

        return inspect_parser

    def cmd_entry(self, cmdargs, config, *args, **kwargs):
        """
        Entry point for the command

        Args:
            self:
            cmdargs:
            config:

        Returns:

        """

        dataset_name = cmdargs.dataset_name
        # description = cmdargs.description
        # statistics = cmdargs.statistics
        # schema = cmdargs.schema
        # metadata = cmdargs.metadata
        preview = False
        split_name = None

        s3_client = ops.OssClient(
            endpoint_url=endpoint_url,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            region_name=region_name,
        )
        db_client = admin.DBClient()
        remote_dataset_list = [
            x.replace("/", "") for x in s3_client.get_dir_list(default_bucket, "")
        ]

        if dataset_name not in remote_dataset_list:
            error_info = "there is no dataset named %s in remote repo" % dataset_name
            raise CLIException(ExistCode.DATASET_NOT_EXIST_REMOTE, error_info)

        if db_client.is_dataset_local_exist(dataset_name):
            dataset_dict = query.get_dataset_info(dataset_name)
        else:
            parquet_key = "/".join([dataset_name, "parquet", "dataset.yaml"])
            dataset_dict = yaml.safe_load(
                s3_client.read_file(default_bucket, parquet_key)
            )

        # if description:
        #     print("dataset description".center(100, "="))
        #     print(dataset_dict["dsdl_meta"]["dataset"]["meta"]["description"])

        # if statistics:
        #     print("dataset statistics".center(100, "="))
        #     print("# " + "basic indicators")
        #     print(
        #         tabulate(
        #             [dataset_dict["statistics"]["dataset_stat"]],
        #             headers="keys",
        #             tablefmt="plain",
        #         )
        #     )
        #
        #     plot_list = dataset_dict["statistics"]["plots"]
        #
        #     for p in plot_list:
        #         if p["type"] == "bar":
        #             name = p["name"]
        #             labels = p["data"]["x_data"]
        #             y_data = p["data"]["y_data"]
        #             data = list(zip(*[x["data"] for x in y_data]))
        #             y_categories = [x["name"] for x in y_data]
        #             y_colors = [x["color"] for x in y_data]
        #             y_label = p["data"]["y_label"]
        #             plot.plt_cmd_bar(
        #                 name, labels, data, y_categories, y_colors, y_label
        #             )
        #
        # if schema:
        #     print("dataset schema".center(100, "="))
        #     schema_dict = dataset_dict["dsdl_meta"]["struct"]
        #     dsdl_version = schema_dict["$dsdl-version"]
        #     fields = schema_dict["Sample"]["$fields"]
        #     optional = schema_dict["Sample"]["$optional"]
        #     schema_tree = Tree(dataset_name)
        #     for k, v in fields.items():
        #         field_str = "%s: %s" % (k, v)
        #         if k in optional:
        #             field_str = field_str + " (Optional)"
        #         schema_tree.add(field_str)
        #
        #     print("# dsdl version: " + dsdl_version)
        #     rprint(schema_tree)

        # if metadata:
        #     print("dataset metadata".center(100, "="))
        #     dataset_meta = dataset_dict["dsdl_meta"]["dataset"]["meta"]
        #     meta_list = []
        #     for k, v in dataset_meta.items():
        #         meta_list.append({"field_name": k, "field_value": v})
        #     print_stdout(tabulate(meta_list, tablefmt='grid', numalign='left'))

        if preview:
            print("Previewing the dataset...")
            from utils.views.view import View

            view = View(dataset_name, info=True)

            from utils.admin import DBClient

            dbcli = DBClient()
            local_exists = dbcli.is_dataset_local_exist(dataset_name)

            if local_exists is True:
                view.view_from_info(split_name)
            else:
                print(f"Dataset {dataset_name} is not exists on local.")

        else:
            dataset_meta = dataset_dict["dsdl_meta"]["dataset"]["meta"]
            meta_list = []
            for k, v in dataset_meta.items():
                meta_list.append({"field_name": k, "field_value": v})
            print_stdout("# dataset info")
            print_stdout(tabulate(meta_list, tablefmt="grid", numalign="left"))

            if "plots" in dataset_dict["statistics"]:
                plot_list = dataset_dict["statistics"]["plots"]

                for p in plot_list:
                    if p["type"] == "bar":
                        name = p["name"]
                        labels = p["data"]["x_data"]
                        y_data = p["data"]["y_data"]
                        data = list(zip(*[x["data"] for x in y_data]))
                        y_categories = [x["name"] for x in y_data]
                        y_colors = [x["color"] for x in y_data]
                        y_label = p["data"]["y_label"]
                        plot.plt_cmd_bar(
                            name, labels, data, y_categories, y_colors, y_label
                        )

            if "tables" in dataset_dict["statistics"]:
                table_list = dataset_dict["statistics"]["tables"]
                for table in table_list:
                    sort_key = list(table["data"][0].keys())[0]
                    print_stdout("# " + table["name"])
                    print_stdout(
                        tabulate(sorted(table["data"], key=lambda d: d[sort_key]),
                                 tablefmt="grid",
                                 headers="keys",
                                 numalign="left",
                                 )
                    )



from .base_geometry import BaseGeometry
import math
import cv2
import numpy as np
from PIL import Image, ImageDraw


class RBBox(BaseGeometry):
    def __init__(self, value, mode):
        if mode == "xywhr":
            self._polygon = None
            self._rbbox = value
        else:
            self._polygon = value
            self._rbbox = None

    @staticmethod
    def rbbox2polygon(value):
        x, y, width, height, angle = value
        cosA, sinA = math.cos(angle), math.sin(angle)

        def _rotate(p_):  # clockwise
            x_, y_ = p_
            x_r = (x_ - x) * cosA - (y_ - y) * sinA + x
            y_r = (x_ - x) * sinA + (y_ - y) * cosA + y
            return [x_r, y_r]

        x_l, x_r, y_t, y_b = x - width / 2, x + width / 2, y - height / 2, y + height / 2
        p_lt, p_lb, p_rt, p_rb = [x_l, y_t], [x_l, y_b], [x_r, y_t], [x_r, y_b]

        return [_rotate(p_lt), _rotate(p_lb), _rotate(p_rb), _rotate(p_rt)]

    def point_for_draw(self, mode: str = "lt"):
        assert mode in ("lb", "lt", "rb", "rt")
        if mode == "lt":
            p = min(self.polygon_value, key=lambda x: x[0] + x[1])
        elif mode == "lb":
            p = min(self.polygon_value, key=lambda x: x[0] - x[1])
        elif mode == "rt":
            p = min(self.polygon_value, key=lambda x: -x[0] + x[1])
        else:  # rb
            p = min(self.polygon_value, key=lambda x: -x[0] - x[1])
        p = [int(_) for _ in p]
        return p

    @staticmethod
    def polygon2rbbox(value):
        res = cv2.minAreaRect(np.array(value).astype(np.int32))
        x, y = res[0]
        width, height = res[1]  # width is "first edge"
        angle = res[2]
        if width < height:
            width, height = height, width
            angle = angle + 90
        angle = 1 - angle / 180 * math.pi
        return [x, y, width, height, angle]

    @property
    def polygon_value(self):
        if self._polygon is None:
            self._polygon = self.rbbox2polygon(self._rbbox)
        return self._polygon

    @property
    def rbbox_value(self):
        if self._rbbox is None:
            self._rbbox = self.polygon2rbbox(self._polygon)
        return self._rbbox

    def visualize(self, image, palette, **kwargs):
        color = (0, 255, 0)
        if "label" in kwargs:
            for label in kwargs["label"].values():
                if label.category_name not in palette:
                    palette[label.category_name] = tuple(np.random.randint(0, 255, size=[3]))

                color = palette[label.category_name]

        poly = Image.new('RGBA', image.size[:2])
        pdraw = ImageDraw.Draw(poly)
        pdraw.polygon([tuple(_) for _ in self.polygon_value], fill=(*color, 127), outline=(*color, 255))
        image.paste(poly, mask=poly)
        del pdraw
        return image

    def __repr__(self):
        x, y, w, h, angle = self.rbbox_value
        x, y, w, h, angle = int(x), int(y), int(w), int(h), int(angle / math.pi * 180)
        return f"[{x}, {y}, {w}, {h}, {angle}°]"

    @property
    def field_key(self):
        return "RotatedBBox"



import os
import signal
import sys
from pathlib import Path
from subprocess import Popen

from loguru import logger

import commons.stdio as stdio
from commons.exceptions import CLIException, ExistCode


class View:
    original_sigint = signal.getsignal(signal.SIGINT)

    def __init__(
        self,
        dataset_name: str = None,
        task: bool = False,
        task_type: str = None,
        info: bool = False,
        number: int = 500,
    ):
        self.dataset_name = dataset_name
        self.task = task
        self.task_type = task_type
        self.info = info
        self.number = number

        if info is True:  # case 'odl-cli info --dataset-name ?? --preview' command
            self.dataset_name = dataset_name
        elif info is False:  # case 'odl-cli studio --dataset-name ?? ...' command
            if (
                task is True
            ):  # case: 'odl-cli studio --dataset-name ?? --task ?? --number ??' command
                self.dataset_name = dataset_name
                self.task_type = task_type
                self.number = number
            else:  # case: 'odl-cli studio dataset_name' command
                self.dataset_name = dataset_name.lower()

        self.view_base_dir = Path(os.path.dirname(__file__))
        self.view_code_name = ".".join([self.dataset_name, "py"])
        self.view_code_abspath = Path.joinpath(
            self.view_base_dir, self.dataset_name, self.view_code_name
        )

    def view_local_dataset(self):
        """
        view the local dataset on webpage.

        returns:
            dataset visualization webpage url.
        """
        stdio.print_stdout(
            f"Processing local dataset {self.dataset_name} visulization..."
        )
        if self.task is True:
            self.view_local_dataset_with_task_type(
                self.dataset_name, self.task_type, self.number
            )
        else:
            self.view_local_dataset_without_task_type(self.dataset_name)

    def view_local_dataset_with_task_type(self, dataset_name, task_type, number):
        """
        view the general dataset on webpage.

        args:
            dataset_name: dataset name.
            task_type: task type name.
            number: number of samples to be viewed.
        returns:
            dataset visualization webpage url.
        """
        viewed = False
        signal.signal(signal.SIGINT, self.ctrl_c_handler)

        view_code_name = "general_app.py"
        view_base_dir = Path(os.path.dirname(__file__))
        view_code_abspath = Path.joinpath(view_base_dir, view_code_name)
        streamlit_cmd = f"streamlit run {view_code_abspath} -- --dataset-name {dataset_name} --task-type {task_type} --number={number}"
        try:
            process = Popen(
                streamlit_cmd,
                shell=True,
            ).wait()
        except KeyboardInterrupt:
            process.send_signal(signal.SIGINT)
            process.wait()
        except CLIException as e:
            stdio.print_stderr(e.message)
            logger.exception(e.message)
            raise CLIException(ExistCode.VIEW_FROM_INSPECT_FAILED, str(e))

        if viewed is False:
            stdio.print_stdout(
                f"Code for Dataset:{self.dataset_name} visualization has not been found, to be done."
            )

    def view_local_dataset_without_task_type(self, dataset_name):
        """
        view the local dataset on webpage.

        args:
            dataset_name: dataset name.
        returns:
            dataset visualization webpage url.
        """
        viewed = False
        signal.signal(signal.SIGINT, self.ctrl_c_handler)
        for dir_name in self.view_base_dir.iterdir():
            if os.path.isdir(dir_name) and dir_name.name == dataset_name:
                for file in dir_name.iterdir():
                    if os.path.exists(self.view_code_abspath):
                        viewed = True
                        stdio.print_stdout(
                            f"Code for Dataset:{dataset_name} visualization has been found."
                        )
                        streamlit_cmd = f"streamlit run {self.view_code_abspath}"
                        try:
                            process = Popen(
                                streamlit_cmd,
                                shell=True,
                            ).wait()
                        except KeyboardInterrupt:
                            process.send_signal(signal.SIGINT)
                            process.wait()
                        except CLIException as e:
                            stdio.print_stderr(e.message)
                            logger.exception(e.message)
                            raise CLIException(
                                ExistCode.VIEW_FROM_INSPECT_FAILED, str(e)
                            )
                        break
                    else:
                        raise CLIException(
                            ExistCode.VIEW_LOCAL_DATASET_FAILED,
                            "Visulization of Dataset:{dataset_name} is to be supported in the future.",
                        )
            else:
                continue

        if viewed is False:
            stdio.print_stdout(
                f"Code for Dataset:{dataset_name} visualization has not been found, to be done."
            )

    def view_remote_dataset(self):
        """
        view the remote dataset on webpage.

        returns:
            dataset visualization webpage url.
        """
        viewed = False
        signal.signal(signal.SIGINT, self.ctrl_c_handler)

        view_code_name = "general_app.py"
        view_base_dir = Path(os.path.dirname(__file__))
        view_code_abspath = Path.joinpath(view_base_dir, view_code_name)
        streamlit_cmd = f"streamlit run {view_code_abspath} -- --dataset-name {self.dataset_name} --task-type {self.task_type} --number={self.number} --remote"
        try:
            process = Popen(
                streamlit_cmd,
                shell=True,
            ).wait()
        except KeyboardInterrupt:
            process.send_signal(signal.SIGINT)
            process.wait()
        except CLIException as e:
            stdio.print_stderr(e.message)
            logger.exception(e.message)
            raise CLIException(ExistCode.VIEW_FROM_INSPECT_FAILED, str(e))

        if viewed is False:
            stdio.print_stdout(
                f"Code for Dataset:{self.dataset_name} visualization has not been found, to be done."
            )

    def view_from_info(self, split_name):
        """
        view the files which get from inspect command on webpage.

        returns:
            dataset visualization webpage url.
        """
        stdio.print_stdout(
            f"Processing local dataset {self.dataset_name} inspect visulization..."
        )

        viewed = False
        signal.signal(signal.SIGINT, self.ctrl_c_handler)

        view_code_name = "view_from_info.py"
        view_base_dir = Path(os.path.dirname(__file__))
        view_code_abspath = Path.joinpath(view_base_dir, view_code_name)
        streamlit_cmd = f"streamlit run {view_code_abspath} -- --dataset-name {self.dataset_name} --split-name {split_name}"
        try:
            process = Popen(
                streamlit_cmd,
                shell=True,
            ).wait()
        except KeyboardInterrupt:
            process.send_signal(signal.SIGINT)
            process.wait()
        except CLIException as e:
            stdio.print_stderr(e.message)
            logger.exception(e.message)
            raise CLIException(ExistCode.VIEW_FROM_INSPECT_FAILED, str(e))

        if viewed is False:
            stdio.print_stdout(
                f"Code for Dataset:{self.dataset_name} visualization has not been found, to be done."
            )

    def ctrl_c_handler(self, signum, frame):
        signal.signal(signal.SIGINT, self.original_sigint)
        stdio.print_stderr("View cancelled")
        logger.exception("View cancelled")
        sys.exit(0)



from boto3.session import Session
import os
from contextlib import contextmanager
from .base import BaseFileReader


class AwsOSSFileReader(BaseFileReader):

    def __init__(self, working_dir, bucket_name, access_key_id, access_key_secret, endpoint, region):
        super().__init__(working_dir)
        self.bucket_name = bucket_name
        self.session = Session(access_key_id, access_key_secret)
        self.s3_client = self.session.client("s3",
                                             endpoint_url=endpoint,
                                             region_name=region,
                                             use_ssl=False)

    @contextmanager
    def load(self, file):
        fp = f"{self.working_dir.strip('/')}/{file.strip('/')}"
        try:
            try:
                data = self.s3_client.get_object(Bucket=self.bucket_name, Key=fp)
                contents = data['Body']
            except Exception as e:
                raise RuntimeError(f"{e}. Failed to read '{fp}' from bucket '{self.bucket_name}'.")
            yield contents
        finally:
            pass



from .field import Field
from ..geometry import BBox, Polygon, PolygonItem, Coord2D, KeyPoints, Text, RBBox, ImageShape, InstanceID
from ..exception import ValidationError
from datetime import date, time, datetime
import math


def validate_list_of_number(value, size_limit, item_type, field_name):
    if type(value) is not list:
        raise ValidationError(f"{field_name} Error: expect list of num, got {value}")
    if len(value) != size_limit:
        raise ValidationError(f"{field_name} Error: expect size of list is {size_limit}, got {len(value)}")
    try:
        return [item_type(item) for item in value]
    except (TypeError, ValueError) as _:
        raise ValidationError(f"{field_name} Error: expect type of list item is float, got {value}")


class CoordField(Field):
    def validate(self, value):
        return validate_list_of_number(value, 2, float, "CoordField")


class Coord3DField(Field):
    def validate(self, value):
        return validate_list_of_number(value, 3, float, "Coord3DField")


class IntervalField(Field):
    def validate(self, value):
        value = validate_list_of_number(value, 2, float, "IntervalField")
        if value[0] > value[1]:
            raise ValidationError(
                f"IntervalField Error: expect |begin| less than or equal to |end|, got {value}"
            )
        return value


class BBoxField(Field):
    def validate(self, value):
        x, y, w, h = validate_list_of_number(value, 4, float, "BBoxField")
        return BBox(x, y, w, h)


class RotatedBBoxField(Field):
    def __init__(self, mode="xywht", measure="radian", *args, **kwargs):
        super(RotatedBBoxField, self).__init__(*args, **kwargs)
        try:
            assert mode in ("xywht", "xyxy")
            self.mode = mode
        except AssertionError as e:
            raise ValidationError("RotateBBox Error: invalid mode, only mode='xywht' or 'xyxy' are permitted.")
        try:
            assert measure in ("radian", "degree")
            self.measure = measure
        except AssertionError as e:
            raise ValidationError("RotateBBox Error: invalid measure, only measure='radian' or 'degree' are permitted.")

    def validate(self, value):
        if self.mode == "xywht":
            value = validate_list_of_number(value, 5, float, "RotateBBoxField")
            if self.measure == "degree":
                value[-1] = value[-1] / 180 * math.pi  # convert to radian
        else:
            value = validate_list_of_number(value, 8, float, "RotateBBoxField")
            value = [value[i:i+2] for i in (0, 2, 4, 6)]
        return RBBox(value, self.mode)


class PolygonField(Field):
    def validate(self, value):
        polygon_lst = []
        for idx, points in enumerate(value):
            for point in points:
                validate_list_of_number(point, 2, float, "PolygonField")
            polygon_lst.append(PolygonItem(points))
        return Polygon(polygon_lst)


class KeypointField(Field):

    def __init__(self, dom, *args, **kwargs):
        super(KeypointField, self).__init__(*args, **kwargs)
        self.dom = dom

    def validate(self, value):
        value = validate_list_of_number(value, len(self.dom), list, "KeypointField")
        keypoints = []
        for class_ind, p in enumerate(value, start=1):
            p = validate_list_of_number(p, 3, float, "KeypointField")
            label = self.dom.get_label(class_ind)
            coord2d = Coord2D(x=p[0], y=p[1], visiable=int(p[2]), label=label)
            keypoints.append(coord2d)

        return KeyPoints(keypoints=keypoints, domain=self.dom)


class LabelField(Field):
    def __init__(self, dom, *args, **kwargs):
        super(LabelField, self).__init__(*args, **kwargs)
        if not isinstance(dom, list):
            self.dom_lst = [dom]
        self.dom_dic = {_.__name__: _ for _ in self.dom_lst}
        self.dom = dom

    def validate(self, value):

        if not isinstance(value, (int, str)):
            raise TypeError(
                f"LabelField Error: invalid class label type. Expected 'int' or 'str' value, got '{value.__class__.__name__}'.")

        if (isinstance(value, int) or (isinstance(value, str) and "::" not in value)):
            if len(self.dom_lst) > 1:
                raise RuntimeError(
                    "LabelField Error: there are more than 1 domains in the struct, you need to specify the label's class domain explicitly.")
            else:
                try:
                    return self.dom.get_label(value)
                except Exception as e:
                    raise RuntimeError(f"LabelField Error: The label '{value}' is not valid. {e}")

        if isinstance(value, str) and "::" in value:
            domain_name, label_name = value.split("::")[:2]

            if domain_name not in self.dom_dic:
                raise RuntimeError(
                    f"LabelField Error: the class domain '{domain_name}' is not provided for the struct."
                )
            else:
                try:
                    return self.dom_dic[domain_name].get_label(label_name)
                except Exception as e:
                    raise RuntimeError(f"LabelField Error: The label '{value}' is not valid. {e}")


class DateField(Field):
    def __init__(self, fmt: str = "", *args, **kwargs):
        super(DateField, self).__init__(*args, **kwargs)
        self.fmt = fmt

    def validate(self, value):
        try:
            if self.fmt == "":
                return date.fromisoformat(value)
            return datetime.strptime(value, self.fmt).date()
        except Exception as e:
            raise RuntimeError(f"DateField Error: {e}")


class TimeField(Field):
    def __init__(self, fmt: str = "", *args, **kwargs):
        super(TimeField, self).__init__(*args, **kwargs)
        self.fmt = fmt

    def validate(self, value):
        try:
            if self.fmt == "":
                return time.fromisoformat(value)
            return datetime.strptime(value, self.fmt).time()
        except Exception as e:
            raise RuntimeError(f"TimeField Error: {e}")


class TextField(Field):
    def validate(self, value):
        try:
            return Text("" + value)
        except TypeError as _:
            raise ValidationError(f"TextField Error: expect str value, got {value}")


class ImageShapeField(Field):
    def __init__(self, mode="hw", *args, **kwargs):
        super(ImageShapeField, self).__init__(*args, **kwargs)
        mode = mode.lower()
        assert mode in ("hw", "wh"), f"the mode of 'ImageShapeField' must be in {('hw', 'wh')}"
        self.mode = mode

    def validate(self, value):
        value = validate_list_of_number(value, 2, int, "ImageShapeField")
        return ImageShape(value=value, mode=self.mode)


class InstanceIDField(Field):
    def validate(self, value):
        try:
            return InstanceID("" + value)
        except TypeError as _:
            raise ValidationError(f"InstanceIDField Error: expect str value, got {value}")



from ...objectio import LocalFileReader
from .base_studio_view import BaseStudioView

try:
    from yaml import CSafeLoader as YAMLSafeLoader
except ImportError:
    from yaml import SafeLoader as YAMLSafeLoader


class LocalStudioView(BaseStudioView):
    def __init__(self, dataset_name, task_type, n=None, shuffle=False):
        super().__init__(dataset_name, task_type, n=n, shuffle=shuffle)

    def _init_file_reader(self):
        return LocalFileReader(self.media_dir)



import warnings

from .utils import *
from dsdl.exception import DefineSyntaxError, ValidationError
from dsdl.warning import DefineSyntaxWarning
from dataclasses import dataclass, field
import re
from typing import Optional


@dataclass()
class EleClass:
    label_value: str
    super_categories: Optional[List[str]] = field(default_factory=list)


class ParserClass:
    def __init__(self, class_name: str, class_value: List[str], skeleton: List = None):
        self.class_name = self._parse_class_name(raw_name=class_name)
        self.class_field = self._parse_class_value(class_value=class_value)
        self.skeleton = self._parse_skeleton(skeleton=skeleton)

    @staticmethod
    def _parse_class_name(raw_name: str) -> str:
        # 判断是否为python保留字
        try:
            check_name_format(raw_name)
        except ValidationError as e:
            raise ValidationError(f"Error with class-dom name, {e}")
        return raw_name

    def _parse_skeleton(self, skeleton: List) -> Optional[List[List[int]]]:
        if not skeleton:
            return None
        else:
            if isinstance(skeleton, list):
                if all(isinstance(item, list) for item in skeleton):
                    for item in skeleton:
                        if not all(
                                isinstance(ele, int) and len(item) == 2 for ele in item
                        ):
                            raise DefineSyntaxError(
                                f"Error in skeleton of {self.class_name}: skeleton must be list of list of int."
                            )
                    return skeleton
                else:
                    raise DefineSyntaxError(
                        f"Error in skeleton of {self.class_name}: skeleton must be list of list of int."
                    )
            else:
                raise DefineSyntaxError(
                    f"Error in skeleton of {self.class_name}: skeleton must be list of list of int."
                )

    def _check_label_name(self, label_name: str) -> str:
        if "." in label_name:
            check_flag = re.search(r"^[\w.]+$", label_name)
            if label_name.startswith("."):
                raise DefineSyntaxError(f"`{label_name}` is not allowed. Label in class-dom can't starts with dot `.`")
            if not check_flag:
                warn_msg = (
                    f"`{label_name}` is not recommended."
                    f" we recommend use alphanumeric letters (a-z, A-Z and 0-9), and underscores (_) "
                    f"for label in class-dom (with hierarchical structure)."
                )
                warnings.warn(
                    warn_msg,
                    DefineSyntaxWarning,
                )
        else:
            check_flag = re.search(r"^[\w\s]+$", label_name)
            if not check_flag:
                warn_msg = (
                    f"`{label_name}` is not recommended."
                    f" we recommend use space, alphanumeric letters (a-z, A-Z and 0-9), and underscores (_) "
                    f"for label in class-dom (without hierarchical structure)."
                )
                warnings.warn(
                    warn_msg,
                    DefineSyntaxWarning,
                )
        return label_name.strip()

    def _parse_class_value(self, class_value: List[str]) -> List[EleClass]:
        class_field = list()
        for value in class_value:
            value = self._check_label_name(value)
            ele_class = EleClass(label_value=value, super_categories=None)
            class_field.append(ele_class)
        return class_field



from typing import Optional
from pydantic import BaseModel


class BaseResponse(BaseModel):
    err_code: int
    err_msg: str

    @staticmethod
    def decorate(target: dict, code: int = 0, msg: str = "success"):
        """
        修饰器，给字典对象添加返回公共头部
        :param target: 要修改的字典
        :param code: 错误码(0为正确)
        :param msg: 错误描述(success为成功)
        :return: 新的字典
        """
        target["err_code"] = code
        target["err_msg"] = msg
        return target


class CheckParam(BaseModel):
    dsdl_yaml: str
    dsdl_library_path: str = "dsdl/dsdl_library"
    output_file: Optional[str] = None
    check_flag: bool = 1





from .base_geometry import BaseGeometry


class InstanceID(BaseGeometry):

    def __init__(self, value):
        self._value = value

    @property
    def value(self):
        return self._value

    def __eq__(self, other):
        return self.value == other.value

    def __repr__(self):
        return f"instance {self.value}"

    @property
    def field_key(self):
        return "InstanceID"



from .field import Field
from ..geometry import Attributes, STRUCT
from ..exception import ValidationError
from ..warning import FieldNotFoundWarning


class StructMetaclass(type):
    def __new__(mcs, name, bases, attributes):
        super_new = super().__new__

        # Also ensure initialization is only performed for subclasses of Model
        # (excluding Model class itself).
        parents = [b for b in bases if isinstance(b, StructMetaclass)]
        if not parents:
            return super_new(mcs, name, bases, attributes)

        required = dict()  # 所有必须填写的field对象
        optional = dict()  # 所有可不填写的field对象
        attr = dict()  # 所有是attribute的field对象
        mappings = dict()  # 所有的field对象
        struct_mappings = dict()  # 所有的Struct对象

        for k, v in attributes.items():
            if isinstance(v, Field):
                mappings[k] = v
                if v.is_attr:
                    attr[k] = v
                if v.is_optional:
                    optional[k] = v
                else:
                    required[k] = v
            elif isinstance(v, Struct):
                struct_mappings[k] = v

        for k in required.keys():
            attributes.pop(k)
        for k in optional.keys():
            attributes.pop(k)
        for k in struct_mappings.keys():
            attributes.pop(k)

        attributes["__required__"] = required
        attributes["__attr__"] = attr
        attributes["__optional__"] = optional
        attributes["__mappings__"] = mappings
        attributes["__struct_mappings__"] = struct_mappings

        new_class = super_new(mcs, name, bases, attributes)
        STRUCT.register(name, new_class)
        return new_class


class Struct(dict, metaclass=StructMetaclass):
    def __init__(self, file_reader=None, **kwargs):

        super().__init__()
        self.attributes = Attributes()
        self.file_reader = file_reader
        self._keys = []
        self._dict_format = None
        if file_reader is not None:  # 说明是在赋值
            for k in self.__required__:
                if k not in kwargs:
                    FieldNotFoundWarning(f"Required field {k} is missing.")
                    continue
                setattr(self, k, kwargs[k])
                if k not in self.__attr__:
                    self._keys.append(k)
        for k in self.__optional__:
            if k in kwargs:
                setattr(self, k, kwargs[k])
                if k not in self.__attr__:
                    self._keys.append(k)
        for k in self.__struct_mappings__:
            if k not in kwargs:
                FieldNotFoundWarning(f"Required struct instance {k} is missing.")
                continue
            setattr(self, k, kwargs[k])
            self._keys.append(k)

        self["$attributes"] = self.attributes
        self._keys.append("$attributes")

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(r"'Model' object has no attribute '%s'" % key)

    def __setattr__(self, key, value):

        if key in self.__mappings__:
            if hasattr(self.__mappings__[key], "set_file_reader"):
                self.__mappings__[key].set_file_reader(self.file_reader)
            try:
                if key in self.__attr__:
                    self.attributes[key] = self.__mappings__[key].validate(value)
                    return
                self[key] = self.__mappings__[key].validate(value)
            except ValidationError as error:
                raise ValidationError(f"Field '{key}' validation error: {error}.")
        elif key in self.__struct_mappings__:
            cls = self.__struct_mappings__[key].__class__
            if not isinstance(value, dict):
                raise ValidationError(
                    f"Struct validation error: {cls.__name__} requires a dict to initiate, but got '{value}'.")
            self[key] = cls(file_reader=self.file_reader, **value)
        else:
            self[key] = value
            return

    def get_mapping(self):
        return self.__mappings__

    def get_struct_mapping(self):
        return self.__struct_mappings__

    def keys(self):
        return tuple(self._keys)

    def convert2dict(self):
        """
        Convert struct instance to a dict like:
             sample = {
                "$image": {"img2": img_obj, "img1":img_obj},
                "$list": {"objects": [{"$image": {"img1": img_obj}, "$bbox": {"box": box_obj}}],
                            "object2": [{"$bbox":{"box": box_obj}}, {"$bbox":{"box": box_obj}}], },
                "$struct": {
                    "struct1": {
                        "$image": {"img3": img_obj, "img4": img_obj}
                    "struct2": {
                        "$image": {"img4": img_obj}}
                    }
                }
            }
        """
        if self._dict_format is None:
            self._dict_format = self._parse_struct(self)
        return self._dict_format

    def extract_field_info(self, field_lst):
        """
        Extract the field info given field list, for example, if field_lst is [bbox, image], the result will be:

            {
                ”image": {
                    "./img2": img_obj,
                    "./img1": img_obj,
                    "./objects/0/img1": img_obj,
                },
                "bbox": {
                    "./objects/0/box": box_obj,
                    "./object2/0/box": box_obj,
                    "./object2/1/box": box_obj
                }
            }
        """
        result_dic = {}
        for field_name in field_lst:
            result_dic[field_name] = self._flatten_sample(f"${field_name}")
        return result_dic

    def _flatten_sample(self, field_name, parse_method=lambda _: _):
        result_dic = {}
        prefix = "."
        self._parse_helper(self.convert2dict(), field_name, result_dic, prefix, parse_method)
        return result_dic

    @classmethod
    def _parse_struct(cls, sample):
        if isinstance(sample, Struct):
            data_item = {}
            field_mapping = sample.get_mapping()  # all fields
            struct_mapping = sample.get_struct_mapping()  # all structs
            for key in sample.keys():
                if key.startswith("$"):  # attributes
                    field_key = key
                    key = key.replace("$", "")
                    if field_key in data_item:
                        data_item[field_key][key] = getattr(sample, field_key)
                    else:
                        data_item[field_key] = {key: getattr(sample, field_key)}
                elif key in field_mapping:  # fields
                    field_obj = getattr(sample, key)
                    field_key = field_mapping[key].extract_key()
                    if field_key in data_item:
                        data_item[field_key][key] = cls._parse_struct(field_obj)
                    else:
                        data_item[field_key] = {key: cls._parse_struct(field_obj)}
                elif key in struct_mapping:  # struct
                    struct_key = "$struct"
                    if struct_key in data_item:
                        data_item[struct_key][key] = cls._parse_struct(getattr(sample, key))
                    else:
                        data_item[struct_key] = {key: cls._parse_struct(getattr(sample, key))}

            return data_item

        elif isinstance(sample, list):
            return [cls._parse_struct(item) for item in sample]
        else:
            return sample

    @classmethod
    def _parse_helper(cls, sample, field_name, result_dic, prefix=".", parse_method=lambda _: _):
        assert field_name not in ("$struct", "$list")

        if isinstance(sample, dict):
            for field_type in sample:
                if field_type == field_name:
                    for key, value in sample[field_type].items():
                        k_ = f"{prefix}/{key}"
                        result_dic[k_] = parse_method(value)
                if field_type == "$list":
                    for key, value in sample[field_type].items():
                        k_ = f"{prefix}/{key}"
                        cls._parse_helper(value, field_name, result_dic, prefix=k_, parse_method=parse_method)
                if field_type == "$struct":
                    for key, value in sample[field_type].items():
                        k_ = f"{prefix}/{key}"
                        cls._parse_helper(value, field_name, result_dic, prefix=k_, parse_method=parse_method)

        elif isinstance(sample, list):
            for id_, item in enumerate(sample):
                k_ = f"{prefix}/{id_}"
                cls._parse_helper(item, field_name, result_dic, prefix=k_, parse_method=parse_method)



import re
import networkx as nx
from typing import Dict, List
from jsonmodels import models, fields, validators
import keyword
from dsdl.exception import ValidationError


TYPES_WITHOUT_PARS = [
    "Bool",
    "Num",
    "Int",
    "Str",
    "Coord",
    "Coord3D",
    "Interval",
    "BBox",
    "Polygon",
    "Image",
    "InstanceMap",
    "Video",
    "Dict",
    "Text",
    "InstanceID",
]
TYPES_TIME = ["Date", "Time"]
TYPES_LABEL = ["Label", "LabelMap", "Keypoint"]
TYPES_LIST = ["List"]
TYPES_IMAGE_SHAPE = ["ImageShape"]
TYPES_ROTATED_BBOX = ["RotatedBBox"]
TYPES_ALL = TYPES_WITHOUT_PARS + TYPES_TIME + TYPES_LABEL + TYPES_LIST + TYPES_IMAGE_SHAPE + TYPES_ROTATED_BBOX


class CheckLogItem(models.Base):
    def_name = fields.StringField(required=True, validators=validators.Enum("class_domain", "struct", "all"))
    yaml = fields.StringField(nullable=True)
    flag = fields.IntField(required=True, validators=validators.Enum(0, 1), default=0)  # 0:error, 1: right
    msg = fields.StringField(nullable=True)


class CheckLog(models.Base):
    def_name = fields.StringField(required=True, validators=validators.Enum("class_domain", "struct", "all"))
    flag = fields.IntField(required=True, validators=validators.Enum(0, 1), default=0)  # 0:error, 1: right
    sub_struct = fields.ListField([CheckLogItem], nullable=True)


def sanitize_variable_name(varstr: str) -> str:
    """
    1. 将`.`替换为`__` 2.将（非字母开头）和（非字母数字及下划线）替换为`_`
    eg. apple.fruit_and_vegetables会转化为apple__fruit_and_vegetables
    """
    temp = varstr.split(".")
    temp = [re.sub("\W|^(?=\d)", "_", i) for i in temp]
    temp = "__".join(temp)
    return temp


def check_name_format(varstr: str):
    if not varstr.isidentifier():
        err_msg = (
            f"`{varstr}` must be a a valid identifier. "
            f"[1. `Struct` name 2. `Class domain` name 3.name of `$field` in `Struct`] "
            f"is considered a valid identifier if "
            f"it only contains alphanumeric letters (a-z) and (0-9), or underscores (_). "
            f"A valid identifier cannot start with a number, or contain any spaces."
        )
        raise ValidationError(f"{err_msg}")
    # 判断是否为python保留字
    if keyword.iskeyword(varstr):
        err_msg = (
            f"`{varstr}` can't be a Python keyword."
            f"check https://docs.python.org/3/reference/lexical_analysis.html#keywords "
            f"for more information."
        )
        raise ValidationError(err_msg)


def rreplace(s, old, new, occurrence):
    """
    从右向左的替换函数，类似replace,不过是反着的
    """
    li = s.rsplit(old, occurrence)
    return new.join(li)


def add_key_value_2_struct_field(field: str, key: str, value) -> str:
    """
    add key, value to field in struct,
    eg: key: optional, value: True, field: NumField(is_attr=True)
    return: NumField(optional=True, is_attr=True)
    """
    p = re.compile(r"[(](.*)[)]", re.S)  # 贪婪匹配
    k_v_list = re.findall(p, field)[0].strip()
    if k_v_list:
        k_v_list = k_v_list.split(",")
    else:
        k_v_list = []
    k_v_list.append(str(key) + "=" + str(value))
    temp = "(" + ", ".join(k_v_list) + ")"
    return field.replace("(" + re.findall(p, field)[0] + ")", temp)


def sort_nx(
    dict_sort_key: Dict[str, List[str]],
) -> List:
    """
    利用有向图对嵌套结构进行排序
    Args:
        dict_sort_key： {当前节点：[父节点],...}
    Returns: 排好序的节点的list，从父到子
    """
    define_graph = nx.DiGraph()
    define_graph.add_nodes_from(dict_sort_key.keys())
    for key, val in dict_sort_key.items():
        for base in val:
            for k in dict_sort_key.keys():
                if k in base:
                    define_graph.add_edge(k, key)
    if not nx.is_directed_acyclic_graph(define_graph):
        raise "define cycle found."
    ordered_keys = list(nx.topological_sort(define_graph))
    return ordered_keys



try:
    from torch.utils.data import Dataset as Dataset_
except ImportError:
    from ..warning import ImportWarning

    ImportWarning("'torch' is not installed.")


    class Dataset_:
        def __len__(self):
            pass

        def __getitem__(self, item):
            pass
from ..types import Struct, StructMetaclass
from ..geometry import STRUCT
from dsdl.dataset.utils import Util
import dsdl.objectio as objectio
from typing import List, Dict, Any, Callable, Optional, Union

try:
    from yaml import CSafeLoader as YAMLSafeLoader
except ImportError:
    from yaml import SafeLoader as YAMLSafeLoader


class Dataset(Dataset_):
    # PALETTE用来存储每个类别的颜色（用于可视化）
    PALETTE = {}

    def __init__(
            self,
            samples: List[Dict[str, Any]],
            sample_type: Union[str, StructMetaclass],
            location_config: dict,
            pipeline: Optional[Callable[[Dict], Dict]] = None,
            global_info_type: Union[str, StructMetaclass] = None,
            global_info: Dict[str, Any] = None
    ):
        self.location_config = location_config
        self.pipeline = pipeline  # 处理样本的函数
        self._samples = samples  # 样本所在的yaml文件路径
        self._global_info = global_info

        if isinstance(sample_type, str):
            sample_type = Util.extract_sample_type(sample_type)
            self.sample_type = STRUCT.get(sample_type)
        elif issubclass(sample_type, Struct):
            self.sample_type = sample_type
        else:
            raise RuntimeError("sample_type must be a string or a Struct class")

        if global_info_type is not None:
            if isinstance(global_info_type, str):
                global_info_type = Util.extract_sample_type(global_info_type)
                self.global_info_type = STRUCT.get(global_info_type)
            elif issubclass(global_info_type, Struct):
                self.global_info_type = global_info_type
            else:
                raise RuntimeError("global_info_type must be a string or a Struct class")
        else:
            self.global_info_type = None

        self.file_reader = self._load_file_reader(location_config)  # 样本的路径配置（如本地路径或是阿里云路径）
        self.global_info = self._load_global_info()
        self.sample_list = self._load_sample()  # 将yaml文件中的样本内容加载到self.sample_list中
        self._has_global_info = self.global_info is not None

    @staticmethod
    def _load_file_reader(config):
        config = config.copy()
        type_ = config.pop("type")
        cls = getattr(objectio, type_)
        try:
            reader = cls(**config)
        except Exception as e:
            print(f"raise exception {e} whe parse the location config {config}.")
            raise e
        return reader

    def has_global_info(self):
        return self._has_global_info

    def _load_global_info(self):
        if self.global_info_type is not None and self._global_info is not None:
            global_info = self.global_info_type(file_reader=self.file_reader, **self._global_info)
            return global_info
        else:
            return None

    def _load_sample(self):
        """
        该函数的作用是将yaml文件中的样本转换为Struct对象，并存储到sample_list列表中
        """
        sample_list = []
        for sample in self._samples:
            sample_list.append(self.sample_type(file_reader=self.file_reader, **sample))
        return sample_list

    def process_sample(self, sample):
        return sample

    def __len__(self):
        return len(self.sample_list)

    def __getitem__(self, idx):
        sample = self.sample_list[idx]
        data = self.process_sample(sample)
        if self.pipeline is not None:
            data = self.pipeline(data)
        return data

    def get_sample_list(self):
        return self.sample_list

    def get_global_info(self):
        return self.global_info

    @property
    def config(self):
        return self.location_config

    @staticmethod
    def format_sample(sample):
        raise NotImplementedError



import io
from PIL import Image
from ..exception import FileReadError
from .utils import bytes_to_numpy
import numpy as np
import cv2

from .base_geometry import BaseGeometry


class InstanceMap(BaseGeometry):
    """
    A Geometry class for instance segmentation map
    """

    def __init__(self, location, file_reader):
        self._loc = location
        self._reader = file_reader

    @property
    def location(self):
        return self._loc

    def to_bypes(self):
        """
        turn InstanceMap object to bytes
        """
        return io.BytesIO(self._reader.read())

    def to_image(self):
        """
        turn InstanceMap object to PIL.Image
        """
        try:
            img = Image.open(self.to_bypes())
        except Exception as e:
            raise FileReadError(f"Failed to convert bytes to an array. {e}") from None
        return img

    def to_array(self):
        """
        turn InstanceMap object to numpy.ndarray
        """
        return bytes_to_numpy(self.to_bypes())

    def visualize(self, image, palette, **kwargs):
        ins_map = self.to_array()
        color_map = np.zeros((ins_map.shape[0], ins_map[1], 3), dtype=np.uint8)
        ins_ids = np.unique(ins_map)

        for ins_id in ins_ids:
            if ins_id == 0:
                continue
            contour_color = tuple(np.random.randint(0, 255, size=[3]))
            this_map = (ins_map == ins_id).astype(np.uint8) * 255
            _, contours, _ = cv2.findContours(this_map, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
            cv2.drawContours(color_map, contours, -1, contour_color, 2)

        overlay = Image.fromarray(color_map).convert("RGBA")
        overlayed = Image.blend(image, overlay, 0.5)
        return overlayed

    def __repr__(self):
        return f"path:{self.location}"

    @property
    def field_key(self):
        return "InstanceMap"



from typing import Set, List, Dict, Optional
from dataclasses import dataclass, field
import re
import warnings
from dsdl.exception import DefineSyntaxError
from dsdl.warning import DefineSyntaxWarning
from collections import defaultdict
from .utils import sort_nx, TYPES_ALL, check_name_format
from dsdl.exception import ValidationError


@dataclass()
class SingleStructParam:
    struct_name: str
    params_dict: Dict[str, str] = field(default_factory=dict)
    parents_struct: List[str] = field(default_factory=list)

    def __post_init__(self):
        """
        按照规定struct和class dom的名字不能是白皮书中已经包含的类型名，如List这些内定的名字
        """
        if self.struct_name in TYPES_ALL:
            raise ValidationError(f"{self.struct_name} is dsdl build-in value name, please rename it. "
                                  f"Build-in value names are: {','.join(TYPES_ALL)}")
        if self.struct_name in [i + 'Field' for i in TYPES_ALL]:
            raise ValidationError(f"{self.struct_name} is dsdl build-in value name, please rename it. "
                                  f"Build-in value names are: {', '.join(TYPES_ALL)}")
        check_name_format(self.struct_name)



@dataclass()
class StructParams:
    params_list: List[SingleStructParam] = field(default_factory=list)


class ParserParam:
    def __init__(self, sample_type, struct_defi, global_info_type=None):
        self.sample_param_map = self._parse_param_sample_type(raw=sample_type)
        if global_info_type:
            self.global_info_param_map = self._parse_param_sample_type(raw=global_info_type)
        else:
            self.global_info_param_map = None
        self.general_param_map = self._get_params(class_defi=struct_defi)

    @staticmethod
    def _parse_param_sample_type(raw: str) -> Optional[SingleStructParam]:
        """
        raw: SceneAndObjectSample[scenedom=SceneDom,objectdom=ObjectDom]
        """
        c_dom = re.findall(r"\[(.*)\]", raw)
        if c_dom:
            data_sample_type = raw.replace("[" + c_dom[0] + "]", "", 1).strip()
            sample_param_map = SingleStructParam(struct_name=data_sample_type)
            # 'SceneAndObjectSample[scenedom=SceneDom,objectdom=[ObjectDom,aaaDom]]'变成
            # ['scenedom=SceneDom', 'objectdom=[ObjectDom,aaaDom]']
            k_v_list = re.split(r",\s*(?![^\[]*\])", c_dom[0])
            k_v_list = [i.strip() for i in k_v_list]
            for param in k_v_list:
                temp = param.split("=")
                sample_param_map.params_dict.update({temp[0].strip(): temp[1].strip()})
            return sample_param_map
        else:
            return None

    def _get_params(self, class_defi: Dict[str, Dict]) -> Optional[Dict[str, SingleStructParam]]:
        self.general_param_map = defaultdict(SingleStructParam)
        ###################################################################################
        # 1。对class_defi循环，先拿到每个struct的params(self.general_param_map), 需要一个单独的循环，因为定义的顺序不一定
        # 以白皮书实例演练：计算机视觉：带场景分类的目标检测为列：
        # 先得到self.general_param_map =
        # {SceneAndObjectSample:{struct_name: SceneAndObjectSample
        #     params_dict: {scenedom: None, objectdom: None}
        #     parents_struct: []},
        # LocalObjectEntry:{struct_name: LocalObjectEntry
        #     cdom: {cdom: None}
        #     parents_struct: []},}
        for define_name, define_value in class_defi.items():
            if "$def" in define_value and define_value["$def"] == "struct":
                struct_params = define_value.get("$params", None)
                if struct_params:
                    temp = SingleStructParam(
                        struct_name=define_name,
                        params_dict={key: None for key in struct_params},
                    )
                    self.general_param_map[define_name] = temp
        ###################################################################################
        # 2。然后填充`SingleStructParam`中的params_dict和parents_struct元素，
        # self.general_param_map =
        # {SceneAndObjectSample:{struct_name: SceneAndObjectSample
        #     params_dict: {scenedom: SceneDom, objectdom: ObjectDom}
        #     parents_struct: []},
        # LocalObjectEntry:{struct_name: LocalObjectEntry
        #     params_dict: {cdom: $objectdom}
        #     parents_struct: [SceneAndObjectSample]},}
        if len(self.general_param_map) == 0:
            return None
        if len(self.general_param_map) == 1:
            temp_param_map = list(self.general_param_map.values())[0]
            temp_name = list(self.general_param_map.keys())[0]
            # 如果需要填的参数在sample-type中：
            if temp_name == self.sample_param_map.struct_name:
                for key in temp_param_map.params_dict.keys():
                    try:
                        self.general_param_map[temp_name].params_dict[
                            key
                        ] = self.sample_param_map.params_dict[key]
                    except KeyError as e:
                        raise DefineSyntaxError(f"miss the params {e} in definition")
            # 如果需要填的参数在global-info-type中：
            elif self.global_info_param_map and temp_name == self.global_info_param_map.struct_name:
                for key in temp_param_map.params_dict.keys():
                    try:
                        self.general_param_map[temp_name].params_dict[
                            key
                        ] = self.global_info_param_map.params_dict[key]
                    except KeyError as e:
                        raise DefineSyntaxError(f"miss the params {e} in definition")
            else:
                warnings.warn(f"parameters in struct {temp_name} is not defined", DefineSyntaxWarning)
        else:
            for define_name, define_value in class_defi.items():
                if "$def" in define_value and define_value["$def"] == "struct":
                    struct_params = define_value.get("$params", None)
                    if struct_params:
                        if define_name == self.sample_param_map.struct_name:
                            for key in self.general_param_map[define_name].params_dict:
                                try:
                                    self.general_param_map[define_name].params_dict[
                                        key
                                    ] = self.sample_param_map.params_dict[key]
                                except KeyError as e:
                                    raise DefineSyntaxError(
                                        f"miss the params {e} in definition"
                                    )
                        # 如果需要填的参数在global-info-type中：
                        elif self.global_info_param_map and define_name == self.global_info_param_map.struct_name:
                            for key in self.general_param_map[define_name].params_dict:
                                try:
                                    self.general_param_map[define_name].params_dict[
                                        key
                                    ] = self.global_info_param_map.params_dict[key]
                                except KeyError as e:
                                    raise DefineSyntaxError(f"miss the params {e} in definition")

                        for raw_type in define_value["$fields"].values():
                            field_type = raw_type.replace(" ", "")
                            # 得到中括号中的 [...] 中的东西，贪婪匹配
                            for structure in self.general_param_map.keys():
                                # 这步不够严谨的，以后需要优化
                                if structure in field_type:
                                    self.general_param_map[
                                        structure
                                    ].parents_struct.append(define_name)
                                    try:
                                        # 如List[LocalObjectEntry[cdom=$objectdom]]会提取到'cdom=$objectdom'
                                        temp = re.findall(
                                            r"%s\[(.*?)\]" % str(structure), field_type
                                        )[0]
                                    except KeyError:
                                        raise DefineSyntaxError(
                                            f"definition error of filed {field_type}"
                                        )
                                    for param in temp.split(","):
                                        param = param.split("=")
                                        if len(param) != 2:
                                            raise DefineSyntaxError(
                                                f"error in params definition {field_type}"
                                            )
                                        key, value = param[0], param[1]
                                        try:
                                            self.general_param_map[
                                                structure
                                            ].params_dict[key] = value
                                        except KeyError as e:
                                            raise DefineSyntaxError(
                                                f"miss the params {e} in definition"
                                            )
                                    break
            ###################################################################################
            # 3。最后利用有向图对struct_name进行排序（父->子），本例中是[SceneAndObjectSample, LocalObjectEntry]
            # sort_param_dict = {SceneAndObjectSample: [], LocalObjectEntry:[SceneAndObjectSample]}
            sort_param_dict = {}
            for key, val in self.general_param_map.items():
                if len(val.parents_struct) > 1:  # 目前只能处理只有一个父节点的情况
                    raise DefineSyntaxError(f"error in definition")
                else:
                    sort_param_dict[key] = val.parents_struct  # {list[str]}
            ordered_keys = sort_nx(sort_param_dict)  # 用有向图排序
            # 对排序后的struct，按顺序填充参数的具体值，最后得到
            # self.general_param_map =
            # {SceneAndObjectSample:{struct_name: SceneAndObjectSample
            #     params_dict: {scenedom: SceneDom, objectdom: ObjectDom}
            #     parents_struct: []},
            # LocalObjectEntry:{struct_name: SceneAndObjectSample
            #     cdom: {cdom: ObjectDom}
            #     parents_struct: [SceneAndObjectSample]},}
            for struct in ordered_keys:
                parent_struct = self.general_param_map[struct].parents_struct
                if not parent_struct:
                    # 因为加了global-info-type以后有两条线了，所以对于没有父类的情况先不处理，最后第4步再统一检查
                    continue
                    # raise DefineSyntaxError(
                    #     f"each struct must have one parent struct, but {struct} have no"
                    # )
                if len(parent_struct) > 1:
                    raise DefineSyntaxError(
                        "each struct must have one parent struct, but can have more than one child struct.\n"
                        f"{struct} have more than one parent struct"
                    )
                parent_struct = parent_struct[0]
                for key, val in self.general_param_map[struct].params_dict.items():
                    if val is None:
                        raise DefineSyntaxError(f"parameter {key} of {struct} must be defined.")
                    if val.startswith("$"):
                        parent_key = val.replace("$", "").strip()
                        self.general_param_map[struct].params_dict[
                            key
                        ] = self.general_param_map[parent_struct].params_dict[
                            parent_key
                        ]
            ###################################################################################
            # 4。最后对self.general_param_map中所有的struct中的参数进行检验，如果有参数没有实例化，就报错
            for struct, single_struct_param in self.general_param_map.items():
                for key, val in single_struct_param.params_dict.items():
                    if val.startswith("$"):
                        raise DefineSyntaxError(f"parameter {val} of {key} in {struct} must be defined.")
        return self.general_param_map

    def validate_params(self, struct_params_field: Set, struct_name: str) -> Set:
        input_params = set(self.general_param_map[struct_name].params_dict.keys())
        if input_params != struct_params_field:
            raise DefineSyntaxError(
                f"error of definition of params {struct_params_field}"
            )
        return struct_params_field



from ...dataset import ImageVisualizePipeline, Util
import os
import json
import random
from ...warning import FieldNotFoundWarning

try:
    from yaml import CSafeLoader as YAMLSafeLoader
except ImportError:
    from yaml import SafeLoader as YAMLSafeLoader
from ..commons import load_samples, TASK_FIELDS
from ...parser import dsdl_parse
from yaml import load as yaml_load
from ...geometry import LABEL, STRUCT, CLASSDOMAIN


class BaseStudioView:
    def __init__(self, dataset_name, task_type, n=None, shuffle=False):
        self.dataset_name = dataset_name
        assert task_type in TASK_FIELDS, f"invalid task, you can only choose in {list(TASK_FIELDS.keys())}"
        self.fields = TASK_FIELDS[task_type]
        self.task_type = task_type
        self.sample_num = n
        self.shuffle = shuffle
        self.yaml_paths, self.media_dir = self.get_yaml_for_cli(dataset_name)
        self.file_reader = self._init_file_reader()
        self._length = None
        self._palette = dict()
        self._ind = 0
        self._generator = self._init_generator()

    def reinit(self):
        self._palette = dict()
        self._ind = 0
        self._generator = self._init_generator()

    def __iter__(self):
        return self

    def __next__(self):
        if self.sample_num is not None:
            if self._ind < self.sample_num:
                return next(self._generator)
            else:
                raise StopIteration
        else:
            return next(self._generator)

    def __len__(self):
        if self._length is None:
            self._length = self._parse_length()
        return self._length

    def _init_generator(self):
        for dsdl_yaml in self.yaml_paths:
            # print(f"Parsing {dsdl_yaml} ...")
            self.clear_registry()
            yaml_info = self.extract_info_from_yml(dsdl_yaml, shuffle=self.shuffle)
            dsdl_py, sample_type, samples, global_info_type, global_info = yaml_info["dsdl_py"], yaml_info[
                "sample_type"], yaml_info["samples"], yaml_info["global_info_type"], yaml_info["global_info"]
            exec(dsdl_py, {})
            sample_type = self.parse_sample_type(sample_type)
            for sample in samples:
                sample = sample_type(file_reader=self.file_reader, **sample)
                sample = ImageVisualizePipeline(sample=sample, palette=self._palette, field_list=self.fields)
                vis_sample = sample.visualize()
                for _, vis_item in vis_sample.items():
                    self._ind += 1
                    yield vis_item

    def _init_file_reader(self):
        raise NotImplementedError

    def _parse_length(self):
        res = 0
        for dsdl_yaml in self.yaml_paths:
            with open(dsdl_yaml, "r") as f:
                dsdl_info = yaml_load(f, Loader=YAMLSafeLoader)['data']
            if "sample-path" not in dsdl_info or dsdl_info["sample-path"] in ("local", "$local"):
                assert "samples" in dsdl_info, f"Key 'samples' is required in {dsdl_yaml}."
                samples = dsdl_info['samples']
            else:
                sample_path = dsdl_info["sample-path"]
                samples = load_samples(dsdl_yaml, sample_path)
            res += len(samples)
        return res

    @staticmethod
    def parse_sample_type(sample_type):
        sample_type = Util.extract_sample_type(sample_type)
        sample_type = STRUCT.get(sample_type)
        return sample_type

    @staticmethod
    def clear_registry():
        LABEL.clear()
        STRUCT.clear()
        CLASSDOMAIN.clear()

    @staticmethod
    def extract_info_from_yml(dsdl_yaml, shuffle=False):
        with open(dsdl_yaml, "r") as f:
            dsdl_info = yaml_load(f, Loader=YAMLSafeLoader)['data']
        sample_type = dsdl_info['sample-type']
        global_info_type = dsdl_info.get("global-info-type", None)
        global_info = None
        if "sample-path" not in dsdl_info or dsdl_info["sample-path"] in ("local", "$local"):
            assert "samples" in dsdl_info, f"Key 'samples' is required in {dsdl_yaml}."
            samples = dsdl_info['samples']
        else:
            sample_path = dsdl_info["sample-path"]
            samples = load_samples(dsdl_yaml, sample_path)
        if global_info_type is not None:
            if "global-info-path" not in dsdl_info:
                assert "global-info" in dsdl_info, f"Key 'global-info' is required in {dsdl_yaml}."
                global_info = dsdl_info["global_info"]
            else:
                global_info_path = dsdl_info["global-info-path"]
                global_info = load_samples(dsdl_yaml, global_info_path, "global-info")[0]

        dsdl_py = dsdl_parse(dsdl_yaml, dsdl_library_path="")
        None if not shuffle else random.shuffle(samples)
        res = {
            "sample_type": sample_type,
            "global_info_type": global_info_type,
            "samples": samples,
            "global_info": global_info,
            "dsdl_py": dsdl_py
        }
        return res

    @staticmethod
    def get_yaml_for_cli(dataset_name):
        SPLIT_PREFIX = "set-"
        config_path = os.path.join(os.path.expanduser("~"), ".dsdl", "dsdl.json")
        with open(config_path, "r") as f:
            storage_info = json.load(f)["storage"]
        dataset_dir = None
        for storage in storage_info.values():
            if "path" not in storage:
                continue
            parent_dir = storage["path"]
            _dataset_dir = os.path.join(parent_dir, dataset_name)
            if os.path.isdir(_dataset_dir):
                dataset_dir = _dataset_dir
                break
        if dataset_dir is None:
            raise RuntimeError(
                f"Dataset '{dataset_name}' doesn't exist local, please download it use command 'odl-cli get'.")
        yml_dir = os.path.join(dataset_dir, "yml")
        if os.path.isdir(yml_dir) and len(os.listdir(yml_dir)) > 0:
            split_dir_names = [str(_) for _ in os.listdir(yml_dir) if
                               os.path.isdir(os.path.join(yml_dir, _)) and str(_).startswith(SPLIT_PREFIX)]
        else:
            FieldNotFoundWarning(f"'{yml_dir}' is not a directory or is empty, please check again.")
            return [], dataset_dir

        split_names = [_[len(SPLIT_PREFIX):] for _ in split_dir_names]
        yaml_names = [f"{_}.yaml" for _ in split_names]
        yaml_paths = [os.path.join(yml_dir, d, y) for d, y in zip(split_dir_names, yaml_names)]

        res = []
        for yml_path in yaml_paths:
            if not os.path.isfile(yml_path):
                FieldNotFoundWarning(f"'{yml_path}' not found, please check again.")
                continue
            res.append(yml_path)

        return res, dataset_dir



from commands.cmdbase import CmdBase
from commands.const import DSDL_CLI_DATASET_NAME
from commons.argument_parser import EnvDefaultVar
from commons.exceptions import CLIException, ExistCode
from loguru import logger
from rich import print as rprint
from rich.pretty import pprint
from tabulate import tabulate
from utils import admin


class Ls(CmdBase):
    """Show all the available datasets and its storage path.

    Args:
        CmdBase (_type_): _description_
    """
    def init_parser(self, subparsers):
        """_summary_

        Args:
            subparsers (_type_): _description_
        """
        
        status_parser = subparsers.add_parser('ls', help = 'show local datasets', example = 'ls.example')
        status_parser.add_argument('-s',
                                   '--skip-header', 
                                   help = 'toggle display for headers',
                                   action = 'store_true',
                                   required = False)
        status_parser.add_argument('-v',
                                   '--verbose',
                                   action = 'count',
                                   default = 0,
                                   help = 'display detailed dataset info',
                                   required = False)
        status_parser.add_argument('--dataset-name',
                                    type = str,
                                    required = False,
                                    help = 'dataset name')
        return status_parser
    
    def get_dataset_df(self, dataset_name = ''):
        db_client = admin.DBClient()
        if dataset_name:
            dataset_path = db_client.get_local_dataset_path(dataset_name)
            if not dataset_path:
                raise CLIException(ExistCode.DATASET_NOT_EXIST_LOCAL,
                                "Dataset `{}` does not exist locally".format(dataset_name))
        dataset_list = db_client.get_sqlite_dict_list('select * from dataset')

        if len(dataset_list) == 0:
            raise CLIException(ExistCode.NO_DATASET_LOCAL,
                               "Local storage has no datasets !")
        # datasplit_list = db_client.get_sqlite_dict_list('select * from split')
        datasplit_join_list = db_client.get_sqlite_dict_list("select * from split where dataset_name='%s'" %(dataset_name))

        return dataset_list, datasplit_join_list
    
    def datalist_trim(self, dataset_list, datasplit_join_list):
        dataset_list_trim = []
        dataset_list_out = []
        # datasplit_list_trim = []
        datasplit_join_list_trim = []
        datasplit_join_list_out = []
        dataset_remove_list = ['created_time','updated_time','dataset_path','label_data', 'media_data']
        _verbose_list = ['label_data', 'media_data']
        datasplit_remove_list = ['created_time','updated_time','label_data', 'media_data']
        
        for idx, elem in enumerate(dataset_list):
            dataset_list[idx]['label_data_exists'] = True if dataset_list[idx]['label_data'] else False
            dataset_list[idx]['media_data_exists'] = True if dataset_list[idx]['media_data'] else False
            dataset_list[idx]['dataset_media_file_bytes'] = str(self.human_readable_size(dataset_list[idx]['dataset_media_file_bytes']))
            dataset_list_trim.append({key:dataset_list[idx][key] for key in dataset_list[idx] if key not in dataset_remove_list})
            dataset_list_out.append({key:dataset_list[idx][key] for key in dataset_list[idx] if key not in _verbose_list})
        # for idx, elem in enumerate(datasplit_list):
        #     datasplit_list_trim.append({key:datasplit_list[idx][key] for key in datasplit_list[idx] if key not in datasplit_verbose_list})
        
        for idx, elem in enumerate(datasplit_join_list):
            datasplit_join_list[idx]['label_data_exists'] = True if datasplit_join_list[idx]['label_data'] else False
            datasplit_join_list[idx]['media_data_exists'] = True if datasplit_join_list[idx]['media_data'] else False
            datasplit_join_list[idx]['split_media_file_bytes'] = str(self.human_readable_size(datasplit_join_list[idx]['split_media_file_bytes']))
            datasplit_join_list_trim.append({key:datasplit_join_list[idx][key] for key in datasplit_join_list[idx] if key not in datasplit_remove_list})
            datasplit_join_list_out.append({key:datasplit_join_list[idx][key] for key in datasplit_join_list[idx] if key not in _verbose_list})
        
        # return dataset_list_trim, datasplit_list_trim, datasplit_join_list_trim
        return dataset_list_trim, dataset_list_out, datasplit_join_list_trim, datasplit_join_list_out
    
    def human_readable_size(self, size:int, decimal_places = 2):
        for unit in ['B', 'KB', 'MB', 'GB', 'TB', 'PB']:
            if size < 1024.0 or unit == 'PB':
                break
            size /= 1024.0
        return f"{size:.{decimal_places}f} {unit}"
    
    
    def cmd_entry(self, cmdargs, config):
        """
        Entry point for the command.

        Args:
            self:
            args:
            config:

        Returns:

        """
        
        __MAPPING = {'dataset_name': 'DatasetName',
                    'task_type': 'TaskType',
                    'storage_name': 'StorageName',
                    'dataset_media_file_num': 'MediaFileNum',
                    'dataset_media_file_bytes': 'MediaFileBytes',
                    'label_data_exists':'LabelDataExists',
                    'media_data_exists':'MediaDataExists',
                    'dataset_path':'DatasetPath',
                    'created_time':'CreatedTime',
                    'updated_time':'UpdatedTime',
                    'split_type':'SplitType',
                    'split_name': 'SplitName',
                    'split_media_file_num':'MediaFileNum',
                    'split_media_file_bytes': 'MediaFileBytes'}

        dataset_name = cmdargs.dataset_name
        dataset_list, datasplit_join_list= self.get_dataset_df(dataset_name)
        dataset_list_trim, dataset_list_out, datasplit_join_list_trim, datasplit_join_list_out = self.datalist_trim(dataset_list, datasplit_join_list)
        
        # no split show dataset table
        if not cmdargs.dataset_name:
        # default display no argument is given
            if not (cmdargs.skip_header or cmdargs.verbose):
                rprint(tabulate(dataset_list_trim, headers=__MAPPING, tablefmt='plain', numalign='left'))
                
            if cmdargs.verbose:
                if cmdargs.verbose >= 1:
                    # verbose and skip head
                    if cmdargs.skip_header:
                        rprint(tabulate(dataset_list_out, tablefmt='plain',numalign='left'))
                    # verbose but not skip head
                    else:
                        rprint(tabulate(dataset_list_out, headers = __MAPPING, tablefmt='plain',numalign='left'))
                        
            if cmdargs.skip_header:
                # not verbose but skip head
                if not cmdargs.verbose:
                    rprint(tabulate(dataset_list_trim, tablefmt='plain', numalign='left'))
        # show split dataset
        else:
            if not (cmdargs.skip_header or cmdargs.verbose):
                rprint(tabulate(datasplit_join_list_trim, headers=__MAPPING, tablefmt='plain',numalign='left'))
                
            if cmdargs.verbose:
                if cmdargs.verbose >= 1:
                    # verbose and skip head
                    if cmdargs.skip_header:
                        rprint(tabulate(datasplit_join_list_out, tablefmt='plain',numalign='left'))
                    # verbose but not skip head
                    else:
                        rprint(tabulate(datasplit_join_list_out, headers = __MAPPING, tablefmt='plain',numalign='left'))
                        
            if cmdargs.skip_header:
                # not verbose but skip head
                if not cmdargs.verbose:
                    rprint(tabulate(datasplit_join_list_trim, tablefmt='plain',numalign='left'))



from .base_dataset import Dataset
from .utils import check_struct, Report


class CheckDataset(Dataset):

    def __init__(self, report: Report, *args, **kwargs):
        self.report = report
        super().__init__(*args, **kwargs)

    def _load_sample(self):
        """
        该函数的作用是将yaml文件中的样本转换为Struct对象，并存储到sample_list列表中
        """
        sample_list = []
        for sample in self._samples:
            struct_instance, report_info = check_struct(self.sample_type, sample, self.file_reader)
            self.report.add_sample_info(report_info)
            if struct_instance is not None:
                sample_list.append(struct_instance)
        return sample_list

    def _load_global_info(self):
        if self.global_info_type is not None and self._global_info is not None:
            global_info, report_info = check_struct(self.global_info_type, self._global_info, self.file_reader)
            self.report.set_global_info(report_info)
            return global_info
        else:
            return None



"""
一个命令的实现例子

Examples:
    $python cli.py example --show tell me the truth
    >> Namespace(show=["'tell", 'me', 'the', "truth'"], command_handler=<bound method Example.cmd_entry of <commands.example.Example object at 0x0000017E6FD1DB40>>)
    >> ["'tell", 'me', 'the', "truth'"]
"""
import os
import re

import yaml
from commands.cmdbase import CmdBase
from commands.const import DEFAULT_LOCAL_STORAGE_PATH, DSDL_CLI_DATASET_NAME
from commons.argument_parser import EnvDefaultVar
from commons.exceptions import CLIException, ExistCode
from commons.stdio import print_stdout
from loguru import logger
from utils import admin, query
from utils.oss_ops import ops
import duckdb

aws_access_key_id = query.aws_access_key_id
aws_secret_access_key = query.aws_secret_access_key
endpoint_url = query.endpoint_url
region_name = query.region_name
default_bucket = query.default_bucket

default_path = DEFAULT_LOCAL_STORAGE_PATH


def query_error_info(error_info):
    try:
        info = str(error_info).split("\n")
        if "^" in info[-1]:
            error_pos = info[-1].index("^")
            index_list = [m.start() for m in re.finditer(' ', info[-2])]
            index = -1
            space = 0
            for n in index_list:
                if n > error_pos:
                    index = n
                    break
                space = n
            std_info = "\n".join([
                "Query syntax error: ", info[-2][space:index], info[-1][space:]
            ])
        else:
            std_info = error_info

        return std_info
    except Exception as e:
        pass


class Select(CmdBase):
    """
    Example command
    """

    def init_parser(self, subparsers):
        """
        Initialize the parser for the command
        document : https://docs.python.org/zh-cn/3/library/argparse.html#

        Args:
            subparsers:

        Returns:

        """
        select_parser = subparsers.add_parser(
            'select',
            help='select data from dataset',
            example="select.example",
            description='select data from dataset',
        )

        select_parser.add_argument(
            "--split",
            type=str,
            required=True,
            help=
            'split name of the dataset, such as train/test/unlabeled or user self-defined split.',
            metavar='')
        select_parser.add_argument(
            "--filter",
            type=str,
            help=
            'filter data according to given conditions, such as "label=\'bird\'"',
            metavar='')
        # select_parser.add_argument("--fields", type=str,
        #                            help='Fields to be selected. All the files will be selected if the arg is not given. Use "," to split fields.',
        #                            metavar='')
        select_parser.add_argument("--limit",
                                   type=int,
                                   help='limit the number of returned records',
                                   metavar='')
        # select_parser.add_argument("--offset", type=int,
        #                            help='Set the number of rows to skip from the beginning of the returned data before presenting the results',
        #                            metavar='')
        # select_parser.add_argument("--random", type=int,
        #                            help='Set the number of random samples from the returned select result',
        #                            metavar='')
        select_parser.add_argument(
            "--export-name",
            type=str,
            help=
            'save the select result as a split and use the given name to name it',
            metavar='')
        select_parser.add_argument(
            "--output",
            type=str,
            help='the path to save the select result as a new split',
            metavar='')
        select_parser.add_argument(
            "dataset_name",
            action=EnvDefaultVar,
            envvar=DSDL_CLI_DATASET_NAME,
            type=str,
            help=
            'dataset name. The arg is optional only when the default dataset name was set by cd command.',
        )
        return select_parser

    def cmd_entry(self, cmdargs, config, *args, **kwargs):
        """
        Entry point for the command

        Args:
            self:
            cmdargs:
            config:

        Returns:

        """
        conf_dict = admin.get_config_dict()

        dataset_name = cmdargs.dataset_name
        split_name = cmdargs.split
        filter = cmdargs.filter
        # fields = '*' if cmdargs.fields is None else cmdargs.fields
        fields = '*'
        limit = cmdargs.limit
        offset = None  # cmdargs.offset
        random = None  # cmdargs.random
        export_name = cmdargs.export_name
        output = cmdargs.output if cmdargs.output else 'default'
        try:
            output_path = conf_dict['storage'][
                cmdargs.output]['path'] if cmdargs.output else default_path
        except Exception as e:
            error_info = str(
                e
            ) + "\n" + "get output path error, output refers to the storage name, the storage path and name are required to be set with config command"
            raise CLIException(ExistCode.STORAGE_NOT_EXIST, error_info)

        db_client = admin.DBClient()
        s3_client = ops.OssClient(aws_access_key_id=aws_access_key_id,
                                  aws_secret_access_key=aws_secret_access_key,
                                  endpoint_url=endpoint_url,
                                  region_name=region_name)
        duck_cursor = duckdb.connect(database=':memory:')
        # if not db_client.is_dataset_local_exist(dataset_name):
        #     print("there is no dataset named %s locally" % dataset_name)
        #     exit()

        # assert (db_client.is_dataset_local_exist(dataset_name)), "there is no dataset named %s locally" % dataset_name

        if not db_client.is_split_local_exist(dataset_name, split_name):
            print_stdout(
                "there is no split named %s of dataset %s locally, search in remote repo"
                % (split_name, dataset_name))

            if not s3_client.is_split_remote_exist(default_bucket,
                                                   dataset_name, split_name):
                reminder = "there is no split named %s of dataset %s neither in remote repo nor local storage" % (
                    split_name, dataset_name)
                logger.info(reminder)
                raise CLIException(ExistCode.DATASET_NOT_EXIST, reminder)
            else:
                yaml_key = '/'.join([dataset_name, 'parquet', 'dataset.yaml'])
                dataset_dict = yaml.safe_load(
                    s3_client.read_file(default_bucket, yaml_key))
        else:
            yaml_path = os.path.join(db_client.get_local_dataset_path(dataset_name), 'parquet/dataset.yaml')
            with open(yaml_path, 'r') as f:
                dataset_dict = yaml.safe_load(f)

        split_reader = query.SplitReader(dataset_name, split_name)
        media_path_field = dataset_dict['dsdl_meta']['struct']['media_field'] if 'media_field' in \
                                                                                 dataset_dict['dsdl_meta'][
                                                                                     'struct'].keys() else 'image'

        # temp solution
        if type(media_path_field) != type([]):
            media_path_field = [media_path_field]

        try:
            df = split_reader.select(select_cols=fields,
                                     filter_cond=filter,
                                     limit=limit,
                                     offset=offset,
                                     samples=random)
        except Exception as e:
            # print_stdout(str(e))
            logger.error(e)
            raise CLIException(ExistCode.QUERY_SYNTAX_ERROR,
                               query_error_info(e))
        print_stdout(df)

        if cmdargs.output is not None and export_name is None:
            print_stdout(
                "If you want to export select result to a directory, --export-name is required"
            )
            exit()

        if export_name is not None:
            media_relative_path = []
            for field in media_path_field:
                try:
                    media_relative_path = media_relative_path + [x[0] for x in duck_cursor.execute(
                        "select %s from df" % field).fetchall()]
                except Exception as e:
                    pass

            sub_split = query.Split(dataset_name, export_name, output_path)
            media_download_flag = 0
            question = 'The split "%s" has already existed. Do you want to replace it? (y/n)' % export_name
            if sub_split.is_local_exist():
                if not input(question) in ("y", "Y"):
                    exit()

            if db_client.is_dataset_local_exist(dataset_name):
                path_info = "the dataset %s registered locally, the split has to been exported to %s" % (
                    dataset_name, sub_split.parquet_path)
                split_db_info = db_client.get_sqlite_dict_list(
                    "select * from split where dataset_name='%s' and split_name='%s'"
                    % (dataset_name, split_name))
                if len(split_db_info) != 0:
                    # print(split_db_info)
                    media_download_flag = split_db_info[0]['media_data']

                if media_download_flag:
                    media_path = [os.path.join(sub_split.dataset_path, x) for x in media_relative_path]
                    media_stat = admin.get_media_stat(media_path)
                    media_num = media_stat['media_num']
                    media_size = media_stat['media_size']

                    split_stat = {
                        'media_num': media_num,
                        'media_size': media_size
                    }

                    stat = split_reader.get_metadata()
                    stat['split_stat'] = split_stat

                    schema = split_reader.get_schema()

                else:
                    media_prefix = "%s/" % dataset_name
                    media_path = [media_prefix + x for x in media_relative_path]
                    media_num = len(set(media_path))
                    media_size = s3_client.get_sum_size(
                        default_bucket, media_path)

                    split_stat = {
                        'media_num': media_num,
                        'media_size': media_size
                    }
                    stat = split_reader.get_metadata()
                    stat['split_stat'] = split_stat

                    schema = split_reader.get_schema()

            else:
                path_info = "the dataset %s does not exist locally, the split has to been exported to given output %s" % (
                    dataset_name, sub_split.parquet_path)

                media_prefix = "%s/" % dataset_name
                media_path = [media_prefix + x for x in media_relative_path]
                media_num = len(set(media_path))
                media_size = s3_client.get_sum_size(default_bucket, media_path)

                split_stat = {'media_num': media_num, 'media_size': media_size}
                stat = split_reader.get_metadata()
                stat['split_stat'] = split_stat

                schema = split_reader.get_schema()

                dataset_media_num = dataset_dict["statistics"]["dataset_stat"][
                    'media_num']
                dataset_media_size = dataset_dict["statistics"][
                    "dataset_stat"]['media_size']

                db_client.register_dataset(dataset_name, output,
                                           sub_split.dataset_path, 0, 0,
                                           dataset_media_num,
                                           dataset_media_size)

            sub_split.save(df, schema, 'user-defined', 1, media_download_flag,
                           stat)
            print_stdout(path_info)

            # to do operations about fields
            # a, b = query.get_parquet_metadata(parquet_path)
            # print(a)
            # print(b)
            #
            # for n in query.get_parquet_schema(parquet_path):
            #     print(n)



from .commons import Util
from .check import check_struct, Report
from .visualizer import ImageVisualizePipeline

__all__ = [
    "Util",
    "check_struct",
    "Report",
    "ImageVisualizePipeline"
]


from .base import BaseFileReader
from contextlib import contextmanager
import os


class CephFileReader(BaseFileReader):

    def __init__(self, working_dir):
        super().__init__(working_dir)
        try:
            import ceph
        except ImportError:
            raise ImportError('Please install ceph to enable CephBackend.')
        self._client = ceph.S3Client()

    @contextmanager
    def load(self, filepath):
        filepath = str(filepath)
        filepath = os.path.join(self.working_dir, filepath)
        value = self._client.Get(filepath)
        try:
            yield memoryview(value)
        finally:
            pass


class PetrelFileReader(BaseFileReader):
    def __init__(self, working_dir, enable_mc=True):
        super().__init__(working_dir)
        try:
            from petrel_client import client
        except ImportError:
            raise ImportError('Please install petrel_client to enable '
                              'PetrelBackend.')
        self._client = client.Client(enable_mc=enable_mc)

    def _format_path(self, filepath: str) -> str:
        """Convert a ``filepath`` to standard format of petrel oss.

        If the ``filepath`` is concatenated by ``os.path.join``, in a Windows
        environment, the ``filepath`` will be the format of
        's3://bucket_name\\image.jpg'. By invoking :meth:`_format_path`, the
        above ``filepath`` will be converted to 's3://bucket_name/image.jpg'.

        Args:
            filepath (str): Path to be formatted.
        """
        return re.sub(r'\\+', '/', filepath)

    def load(self, filepath):
        filepath = self._format_path(filepath)
        value = self._client.Get(filepath)
        try:
            yield memoryview(value)
        finally:
            pass
